import numpy as np
from mapel.elections.distances.lp import solve_rand_approx_pav
from mapel.elections.features.scores import get_score, get_dissat


# NEW LP
def get_rand_approx_pav_score(election, committee_size=1):
    if election.fake:
        return {'value': None}

    W = [1 / (i + 1) for i in range(election.num_candidates)]

    C = np.zeros([election.num_voters, election.num_candidates])
    for i, vote in enumerate(election.votes):
        for j, c in enumerate(vote):
            C[i][c] = j
    # is C equivalent to potes?

    return solve_rand_approx_pav(election, committee_size, W, C)


# GREEDY
def get_greedy_approx_cc_score(election, committee_size=1):
    return get_greedy_approx_score(election, 'cc', committee_size=committee_size)


def get_greedy_approx_hb_score(election, committee_size=1):
    return get_greedy_approx_score(election, 'hb', committee_size=committee_size)


def get_greedy_approx_pav_score(election, committee_size=1):
    return get_greedy_approx_score(election, 'pav', committee_size=committee_size)


def get_greedy_approx_score(election, rule, committee_size=1):
    if election.fake:
        return {'value': None, 'dissat': None}
    winners = get_winners_approx_greedy(election, committee_size, rule)
    return {'value': get_score(election, winners, rule),
            'dissat': get_dissat(election, winners, rule)}


def get_winners_approx_greedy(election, committee_size, rule):
    """ universal function """

    owa_vector, scoring_vector = get_vectors(election, rule, committee_size)

    if rule == 'hb':
        owa_vector = [1/(i+1) for i in range(election.num_candidates)]
        scoring_vector = [election.num_candidates - i - 1 for i in range(election.num_candidates)]
    elif rule == 'pav':
        owa_vector = [1/(i+1) for i in range(election.num_candidates)]
        scoring_vector = [0] * election.num_candidates
        for i in range(committee_size):
            scoring_vector[i] = 1
    elif rule == 'cc':
        owa_vector = [0] * election.num_candidates
        owa_vector[0] = 1
        scoring_vector = [election.num_candidates - i - 1 for i in range(election.num_candidates)]

    def simple(active, vote, target, owa_vector, scoring_vector):
        income = 0
        ctr = 0
        for x in range(len(vote)):
            if not active[vote[x]] or target == vote[x]:
                income += owa_vector[ctr] * scoring_vector[x]
                ctr += 1
        return income


    winners = []
    voter_sat = [0 for _ in range(election.num_voters)]
    active = [True for _ in range(election.num_candidates)]

    for Z in range(committee_size):

        points = [0 for _ in range(election.num_candidates)]
        income = np.zeros([election.num_voters, election.num_candidates], dtype=float)

        # Compute points
        for i in range(election.num_voters):
            # ctr = 1.
            for j in range(election.num_candidates):
                if active[election.votes[i][j]]:
                    value = simple(active, election.votes[i], election.votes[i][j],
                                   owa_vector, scoring_vector)
                    points[election.votes[i][j]] += value - voter_sat[i]
                    income[i][election.votes[i][j]] = value
                # else:
                    # ctr += 1.

        winner_id = -1
        winner_value = -1

        # Select winner
        # random_order = np.random.permutation(election.num_candidates)
        random_order = [i for i in range(election.num_candidates)]
        for i in random_order:
            if active[i] and points[i] > winner_value:
                winner_value = points[i]
                winner_id = i

        # Update voter satisfaction
        for i in range(election.num_voters):
            voter_sat[i] = income[i][winner_id]

        winners.append(winner_id)
        active[winner_id] = False

    return winners


# REMOVAL
def get_removal_approx_cc_score(election, committee_size=1):
    return get_removal_approx_score(election, "cc", committee_size=committee_size)


def get_removal_approx_hb_score(election, committee_size=1):
    return get_removal_approx_score(election, "hb", committee_size=committee_size)


def get_removal_approx_pav_score(election, committee_size=1):
    return get_removal_approx_score(election, "pav", committee_size=committee_size)


def get_removal_approx_score(election, rule, committee_size=1):
    if election.fake:
        return {'value': None, 'dissat': None}
    winners = get_winners_approx_removal(election, committee_size, rule)

    return {'value': get_score(election, winners, rule),
            'dissat': get_dissat(election, winners, rule)}


def get_winners_approx_removal(election, committee_size, rule):

    owa_vector, scoring_vector = get_vectors(election, rule, committee_size)

    def simple(active, vote, target, owa_vector, scoring_vector):
        income = 0
        ctr = 0
        for x in range(len(vote)):
            if active[vote[x]] and target != vote[x]:
                income += owa_vector[ctr] * scoring_vector[x]
                ctr += 1
        return income

    num_voters = election.num_voters
    num_candidates = election.num_candidates
    votes = election.votes
    removed = 0
    active = [True for _ in range(num_candidates)]

    # PRECOMPUTE
    starting_voter_sat = 0
    for i, x in enumerate(owa_vector):
        starting_voter_sat += scoring_vector[i] * x

    voter_sat = [starting_voter_sat for _ in range(election.num_voters)]

    # STANDARD
    for Z in range(num_candidates - committee_size):

        points = [0. for _ in range(num_candidates)]
        income = np.zeros([election.num_voters, election.num_candidates], dtype=float)

        for i in range(num_voters):
            for j in range(num_candidates):
                if active[votes[i][j]]:

                    value = simple(active, election.votes[i], election.votes[i][j],
                                   owa_vector, scoring_vector)
                    points[election.votes[i][j]] += voter_sat[i] - value
                    income[i][election.votes[i][j]] = value

        loser_id = -1
        loser_value = 9999999
        for i in range(num_candidates):
            if (active[i]) and (0 <= points[i] <= loser_value):
                loser_value = points[i]
                loser_id = i

        active[loser_id] = False
        removed += 1

        # Update voter satisfaction
        for i in range(election.num_voters):
            voter_sat[i] = income[i][loser_id]

    winners = []
    for i in range(num_candidates):
        if active[i]:
            winners.append(i)

    return winners



def get_vectors(election, rule, committee_size):

    if rule == 'hb':
        owa_vector = [1/(i+1) for i in range(election.num_candidates)]
        scoring_vector = [election.num_candidates - i - 1 for i in range(election.num_candidates)]
    elif rule == 'pav':
        owa_vector = [1/(i+1) for i in range(election.num_candidates)]
        scoring_vector = [0] * election.num_candidates
        for i in range(committee_size):
            scoring_vector[i] = 1
    elif rule == 'cc':
        owa_vector = [0] * election.num_candidates
        owa_vector[0] = 1
        scoring_vector = [election.num_candidates - i - 1 for i in range(election.num_candidates)]

    return owa_vector, scoring_vector





import scipy.special
import numpy as np

from mapel.elections.features.scores import get_cc_score, get_cc_dissat


def get_ranging_cc_score(election, committee_size=1):
    if election.fake:
        return {'value': None, 'dissat': None}

    x = election.num_candidates * scipy.special.lambertw(committee_size).real / committee_size

    best_score = 0
    best_dissat = 0
    for threshold in range(1, int(x)):
        score, dissat = get_algorithm_p_committee(election, committee_size, x)
        if score > best_score:
            best_score = score
            best_dissat = dissat

    return {'value': best_score, 'dissat': best_dissat}


def get_algorithm_p_committee(election, committee_size, x):

    winners = set()

    active = [True for _ in range(election.num_voters)]

    for i in range(committee_size):
        tops = np.zeros([election.num_candidates])
        for v, vote in enumerate(election.votes):
            if active[v]:
                for c in range(int(x)):
                    tops[vote[c]] += 1

        winner_id = np.argmax(tops)
        winners.add(winner_id)
        for v, vote in enumerate(election.votes):
            if active[v]:
                for c in range(int(x)):
                    if winner_id == vote[c]:
                        active[v] = False
                        break

    return get_cc_score(election, winners), get_cc_dissat(election, winners)




import numpy as np
import numpy.linalg as la
try:
    import pandas as pd
except:
    pass


def min_dim(election):
    if election.fake:
        return None

    election.set_default_object_type('vote')
    election.compute_distances(distance_id='swap')

    df = election.distances['vote']
    df = pd.DataFrame(df, columns=[str(i) for i in range(election.num_voters)])

    # get some distance matrix
    # df = pd.read_csv("tmp.csv")
    A = df.values.T.astype(float)
    # square it
    A = A ** 2

    # centering matrix
    n = A.shape[0]
    J_c = 1. / n * (np.eye(n) - 1 + (n - 1) * np.eye(n))

    # perform double centering
    B = -0.5 * (J_c.dot(A)).dot(J_c)

    # find eigenvalues and eigenvectors
    eigen_val = la.eig(B)[0]
    eigen_vec = la.eig(B)[1].T

    num_zeros = 0
    for i, _ in enumerate(eigen_val):
        if eigen_val[i] > -0.00001 and eigen_val[i] < 0.00001:
            num_zeros += 1

    return election.num_voters-num_zeros


from random import *

import numpy as np
from mapel.core.features.mallows import phi_from_normphi, mallows_votes
from scipy.special import binom


### MATRICES ###

# WALSH

def f(i, j):
    if i < 0: return 0
    return (1.0 / (2 ** (i + j))) * binom(i + j, i)


def probW(m, i, t):
    # probability that c_i is ranked process_id among m candidates
    return 0.5 * f(i - 1, m - t - (i - 1)) + 0.5 * f(i - t, m - i)


# RANDOM CONITZER

def random_conitzer(C):
    # generate a random vote from the Conitzer culture_id for axis
    # C[0], ..., C[m-1]
    m = len(C)
    center = np.random.randint(0, m)
    left = center
    right = center
    vote = [C[center]]
    for i in range(m - 1):
        L = False
        R = False

        if left > 0 and right < m - 1:
            if random() < 0.5:
                L = True
            else:
                R = True
        elif left > 0:
            L = True
        else:
            R = True

        if L:
            left -= 1
            vote.append(C[left])
        else:
            right += 1
            vote.append(C[right])

    return vote


# CONITZER

def g(m, i, j):
    if i > j: return 0
    if i == j: return 1.0 / m
    if i == 1 and j < m: return g(m, 1, j - 1) + 0.5 * g(m, 2, j)
    if j == m and i > 1: return g(m, i + 1, m) + 0.5 * g(m, i, m - 1)
    if i == 1 and j == m: return 1.0
    return 1.0 / m


#  return 0.5*g(m,i+1,j) + 0.5*g(m,i,j-1)


def probC(m, i, t):
    # probability that c_i is ranked process_id among m candidates
    p = 0.0
    if t == 1: return 1.0 / m

    if i - (t - 1) > 1:
        p += 0.5 * g(m, i - (t - 1), i - 1)
    elif i - (t - 1) == 1:
        p += g(m, i - (t - 1), i - 1)

    if i + (t - 1) < m:
        p += 0.5 * g(m, i + 1, i + (t - 1))
    elif i + (t - 1) == m:
        p += g(m, i + 1, i + (t - 1))

    return p


PRECISION = 1000
DIGITS = 4


def get_conitzer_matrix(m):
    return get_conitzer_vectors(m).transpose()


def get_walsh_matrix(m):
    return get_walsh_vectors(m).transpose()


def get_conitzer_vectors(m):
    P = np.zeros([m, m])
    for i in range(m):
        for j in range(m):
            P[i][j] = probC(m, i + 1, j + 1)
    return P


def simconitzer(m):
    P = [[0] * m for _ in range(m)]
    T = 100000

    C = list(range(m))
    for t in range(T):
        if t % 10000 == 0: print(t)
        v = random_conitzer(C)
        for i in range(m):
            P[v[i]][i] += 1

    for j in range(m):
        for i in range(m):
            P[i][j] = str(int(PRECISION * (P[i][j] / T))).rjust(DIGITS)
    return P


def get_walsh_vectors(m):
    P = np.zeros([m, m])
    for i in range(m):
        for t in range(m):
            P[i][t] = probW(m, i + 1, t + 1)
    return P


########  MALLOWS SP  ########
def generate_conitzer_mallows_votes(num_voters, num_candidates, params):
    params['phi'] = phi_from_normphi(num_candidates, normphi=params['normphi'])

    votes = generate_ordinal_sp_conitzer_votes(num_voters=num_voters, num_candidates=num_candidates)

    votes = mallows_votes(votes, params['phi'])

    return votes


def generate_walsh_mallows_votes(num_voters, num_candidates, params):
    params['phi'] = phi_from_normphi(num_candidates, normphi=params['normphi'])

    votes = generate_ordinal_sp_walsh_votes(num_voters=num_voters, num_candidates=num_candidates)

    votes = mallows_votes(votes, params['phi'])

    return votes


import numpy as np
import math
from numpy import linalg
import os
import csv

####################################################################################################
# Approval Euclidean Election Models
####################################################################################################

def get_range(params):
    if params['p_dist'] == 'beta':
        return np.random.beta(params['a'], params['b'])
    elif params['p_dist'] == 'uniform':
        return np.random.uniform(low=params['a'], high=params['b'])
    else:
        return params['p_dist']


def generate_approval_vcr_votes(num_voters: int = None,
                                num_candidates: int = None,
                                dim: int = None,
                                space: str = None,
                                params=None) -> list:

    votes = [set() for _ in range(num_voters)]

    name = f'{dim}d_{space}'

    voters = np.array([get_rand(name) for _ in range(num_voters)])
    candidates = np.array([get_rand(name) for _ in range(num_candidates)])

    v_range = np.array([get_range(params) for _ in range(num_voters)])
    c_range = np.array([get_range(params) for _ in range(num_candidates)])

    for v in range(num_voters):
        for c in range(num_candidates):
            if v_range[v] + c_range[c] >= np.linalg.norm(voters[v] - candidates[c],
                                                         ord=dim):
                votes[v].add(c)

    return votes


def generate_approval_euclidean_votes(num_voters: int = None,
                                      num_candidates: int = None,
                                      dim: int = 2,
                                      space: str = 'uniform',
                                      radius: float = 0,
                                      **kwargs) -> list:
    """
    Generates approval votes from euclidean model.

        Parameters
        ----------
        num_voters : int
            Number of voters.
        num_candidates : int
            Number of candidates.
        dim : int
            Number of dimensions.
        space : int
            Type of distribution for voters and candidates.
            {'uniform', 'gaussian', 'sphere'}
        radius : float
            The radius.

        Returns
        -------
        list[set]
            Approval votes.

    """

    votes = [set() for _ in range(num_voters)]

    if space == 'uniform':
        voters = np.random.rand(num_voters, dim)
        candidates = np.random.rand(num_candidates, dim)
    elif space == 'gaussian':
        voters = np.random.normal(loc=0.5, scale=0.15, size=(num_voters, dim))
        candidates = np.random.normal(loc=0.5, scale=0.15, size=(num_candidates, dim))
    elif space == 'sphere':
        voters = np.array([list(random_sphere(dim)[0]) for _ in range(num_voters)])
        candidates = np.array([list(random_sphere(dim)[0]) for _ in range(num_candidates)])

    for v in range(num_voters):
        for c in range(num_candidates):
            if radius >= np.linalg.norm(voters[v] - candidates[c]):
                votes[v].add(c)

    return votes

####################################################################################################
# Party Euclidean Election Models
####################################################################################################
def generate_1d_gaussian_party(num_voters=None, num_candidates=None, params=None):
    if params is None:
        params = {}
    if 'num_winners' not in params:
        params['num_winners'] = 1

    voters = [[] for _ in range(num_voters)]
    candidates = [[] for _ in range(num_candidates)]

    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    for j in range(params['num_parties']):
        for w in range(params['num_winners']):
            _id = j * params['num_winners'] + w
            candidates[_id] = [np.random.normal(params['party'][j][0], params['var'])]

    _min = min(candidates)[0]
    _max = max(candidates)[0]

    shift = [np.random.random() / 2. - 1 / 4.]
    for j in range(num_voters):
        voters[j] = [np.random.random() * (_max - _min) + _min + shift[0]]

    for j in range(num_voters):
        for k in range(num_candidates):
            votes[j][k] = k
            distances[j][k] = np.linalg.norm(voters[j] - candidates[k], ord=1)

        votes[j] = [x for _, x in sorted(zip(distances[j], votes[j]))]

    return votes


def generate_2d_gaussian_party(num_voters=None, num_candidates=None, params=None):
    if params is None:
        params = {}
    if 'num_winners' not in params:
        params['num_winners'] = 1

    voters = [[] for _ in range(num_voters)]
    candidates = [[] for _ in range(num_candidates)]

    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    for j in range(params['num_parties']):
        for w in range(params['num_winners']):
            _id = j * params['num_winners'] + w
            candidates[_id] = [np.random.normal(params['party'][j][0][0], params['var']),
                               np.random.normal(params['party'][j][0][1], params['var'])]

    def column(matrix, i):
        return [row[i] for row in matrix]

    x_min = min(column(candidates, 0))
    x_max = max(column(candidates, 0))
    y_min = min(column(candidates, 1))
    y_max = max(column(candidates, 1))

    shift = [np.random.random() / 2. - 1 / 4., np.random.random() / 2. - 1 / 4.]
    for j in range(num_voters):
        voters[j] = [np.random.random() * (x_max - x_min) + x_min + shift[0],
                     np.random.random() * (y_max - y_min) + y_min + shift[1]]

    for j in range(num_voters):
        for k in range(num_candidates):
            votes[j][k] = k
            distances[j][k] = np.linalg.norm(voters[j] - candidates[k], ord=2)

        votes[j] = [x for _, x in sorted(zip(distances[j], votes[j]))]

    return votes


####################################################################################################
# Ordinal Euclidean Election Models
####################################################################################################
def store_ideal_points(points, file_name, experiment_id):
    path = os.path.join(os.getcwd(), "election", experiment_id, "elections", f'{file_name}.csv')
    with open(path, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(["x", "y"])
        for point in points:
            writer.writerow([point[0], point[1]])

def generate_ordinal_euclidean_votes(model: str = 'euclidean', num_voters: int = None,
                                     num_candidates: int = None,
                                     dim: int = 2, space: str = 'uniform',
                                     aggregated='True',
                                     ele_id=None,
                                     exp_id=None) -> np.ndarray:

    voters = np.zeros([num_voters, dim])
    candidates = np.zeros([num_candidates, dim])
    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    if model == 'euclidean':
        if space == 'uniform':
            voters = np.random.rand(num_voters, dim)
            candidates = np.random.rand(num_candidates, dim)
        elif space == 'gaussian':
            voters = np.random.normal(loc=0.5, scale=0.15, size=(num_voters, dim))
            candidates = np.random.normal(loc=0.5, scale=0.15, size=(num_candidates, dim))
        elif space == 'sphere':
            voters = np.array([list(random_sphere(dim)[0]) for _ in range(num_voters)])
            candidates = np.array([list(random_sphere(dim)[0]) for _ in range(num_candidates)])
    else:
        for v in range(num_voters):
            voters[v] = get_rand(model)
        # voters = sorted(voters)

        for v in range(num_candidates):
            candidates[v] = get_rand(model)
        # candidates = sorted(candidates)

    for v in range(num_voters):
        for c in range(num_candidates):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(voters[v] - candidates[c], ord=dim)

        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    if not aggregated:
        store_ideal_points(voters, f'{ele_id}_voters', exp_id)
        store_ideal_points(candidates, f'{ele_id}_candidates', exp_id)

    return votes


def generate_elections_2d_grid(num_voters=None, num_candidates=None):
    voters = np.zeros([num_voters, 2])
    candidates = []

    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    for j in range(num_voters):
        voters[j] = get_rand('2d_square')
    voters = sorted(voters)

    sq = int(num_candidates ** 0.5)
    d = 1. / sq

    for i in range(sq):
        for j in range(sq):
            x = d / 2. + d * i
            y = d / 2. + d * j
            point = [x, y]
            candidates.append(point)

    for j in range(num_voters):
        for k in range(num_candidates):
            votes[j][k] = k
            distances[j][k] = np.linalg.norm(voters[j] - candidates[k], ord=2)

        votes[j] = [x for _, x in sorted(zip(distances[j], votes[j]))]

    return votes


# AUXILIARY
def random_ball(dimension, num_points=1, radius=1):
    random_directions = np.random.normal(size=(dimension, num_points))
    random_directions /= linalg.norm(random_directions, axis=0)
    random_radii = np.random.random(num_points) ** (1 / dimension)
    x = radius * (random_directions * random_radii).T
    return x


def random_sphere(dimension, num_points=1, radius=1):
    random_directions = np.random.normal(size=(dimension, num_points))
    random_directions /= linalg.norm(random_directions, axis=0)
    random_radii = 1.
    return radius * (random_directions * random_radii).T


def get_rand(model: str, cat: str = "voters") -> list:
    """ generate random values """

    point = [0]
    if model in {"1d_uniform",  "1d_interval"}:
        return np.random.rand()
    elif model in {'1d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=1)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=1)
    elif model in {"1d_gaussian"}:
        point = np.random.normal(0.5, 0.15)
        while point > 1 or point < 0:
            point = np.random.normal(0.5, 0.15)
    elif model == "1d_one_sided_triangle":
        point = np.random.uniform(0, 1) ** 0.5
    elif model == "1d_full_triangle":
        point = np.random.choice([np.random.uniform(0, 1) ** 0.5, 2 - np.random.uniform(0, 1) ** 0.5])
    elif model == "1d_two_party":
        point = np.random.choice([np.random.uniform(0, 1), np.random.uniform(2, 3)])
    elif model in {"2d_disc", "2d_range_disc"}:
        phi = 2.0 * 180.0 * np.random.random()
        radius = math.sqrt(np.random.random()) * 0.5
        point = [0.5 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
    elif model == "2d_range_overlapping":
        phi = 2.0 * 180.0 * np.random.random()
        radius = math.sqrt(np.random.random()) * 0.5
        if cat == "voters":
            point = [0.25 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
        elif cat == "candidates":
            point = [0.75 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
    elif model in {"2d_square", "2d_uniform"}:
        point = [np.random.random(), np.random.random()]
    elif model in {'2d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=2)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=2)
    elif model == "2d_sphere":
        alpha = 2 * math.pi * np.random.random()
        x = 1. * math.cos(alpha)
        y = 1. * math.sin(alpha)
        point = [x, y]
    elif model == "2d_ball":
        dim = 2
        point = list(random_ball(dim, radius=0.5)[0])
    elif model in ["2d_gaussian", "2d_range_gaussian"]:
        point = [np.random.normal(0.5, 0.15), np.random.normal(0.5, 0.15)]
        while np.linalg.norm(point - np.array([0.5, 0.5]), ord=2) > 0.5:
            point = [np.random.normal(0.5, 0.15), np.random.normal(0.5, 0.15)]
    elif model in ["2d_range_fourgau"]:
        r = np.random.randint(1, 4)
        size = 0.06
        if r == 1:
            point = [np.random.normal(0.25, size), np.random.normal(0.5, size)]
        if r == 2:
            point = [np.random.normal(0.5, size), np.random.normal(0.75, size)]
        if r == 3:
            point = [np.random.normal(0.75, size), np.random.normal(0.5, size)]
        if r == 4:
            point = [np.random.normal(0.5, size), np.random.normal(0.25, size)]
    elif model in ["3d_cube", "3d_uniform"]:
        point = [np.random.random(), np.random.random(), np.random.random()]
    elif model in {'3d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=3)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=3)
    elif model == "4d_cube":
        dim = 4
        point = [np.random.random() for _ in range(dim)]
    elif model == "5d_cube":
        dim = 5
        point = [np.random.random() for _ in range(dim)]
    elif model == "10d_cube":
        dim = 10
        point = [np.random.random() for _ in range(dim)]
    elif model == "20d_cube":
        dim = 20
        point = [np.random.random() for _ in range(dim)]
    elif model == "3d_sphere":
        dim = 3
        point = list(random_sphere(dim)[0])
    elif model == "4d_sphere":
        dim = 4
        point = list(random_sphere(dim)[0])
    elif model == "5d_sphere":
        dim = 5
        point = list(random_sphere(dim)[0])
    elif model == '2d_overlapping_squares':
        if cat == 'voters':
            return [np.random.uniform([0, 0.75]), np.random.uniform([0., 0.75])]
        elif cat == 'candidates':
            return [np.random.uniform([0.25, 1]), np.random.uniform([0.25, 1])]
    elif model == '2d_asymmetric_gaussians':
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=2)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=2)
    else:
        logging.warning(f'No such culture id: {model}')
        point = [0, 0]
    return point



#!/usr/bin/env python
import os
from abc import ABC

from mapel.elections.objects.ElectionExperiment import ElectionExperiment
import mapel.elections.cultures_ as cultures
import mapel.elections.features_ as features
import mapel.elections.distances_ as distances

try:
    from sklearn.manifold import MDS
    from sklearn.manifold import TSNE
    from sklearn.manifold import SpectralEmbedding
    from sklearn.manifold import LocallyLinearEmbedding
    from sklearn.manifold import Isomap
except ImportError as error:
    MDS = None
    TSNE = None
    SpectralEmbedding = None
    LocallyLinearEmbedding = None
    Isomap = None
    print(error)


class OrdinalElectionExperiment(ElectionExperiment, ABC):
    """Abstract set of elections."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_culture(self, name, function):
        cultures.add_ordinal_culture(name, function)

    def add_feature(self, name, function):
        features.add_ordinal_feature(name, function)

    def add_distance(self, name, function):
        distances.add_ordinal_distance(name, function)

    def add_folders_to_experiment(self) -> None:

        dirs = ["experiments", "images", "trash"]
        for dir in dirs:
            if not os.path.isdir(dir):
                os.mkdir(os.path.join(os.getcwd(), dir))

        if not os.path.isdir(os.path.join(os.getcwd(), "experiments", self.experiment_id)):
            os.mkdir(os.path.join(os.getcwd(), "experiments", self.experiment_id))

        list_of_folders = ['distances',
                           'features',
                           'coordinates',
                           'elections',
                           'matrices']

        for folder_name in list_of_folders:
            if not os.path.isdir(os.path.join(os.getcwd(), "experiments",
                                              self.experiment_id, folder_name)):
                os.mkdir(os.path.join(os.getcwd(), "experiments",
                                      self.experiment_id, folder_name))

        path = os.path.join(os.getcwd(), "experiments", self.experiment_id, "map.csv")
        if not os.path.exists(path):

            with open(path, 'w') as file_csv:
                file_csv.write(
                    "size;num_candidates;num_voters;culture_id;params;color;alpha;"
                    "family_id;label;marker;show\n")
                file_csv.write("3;10;100;ic;{};black;1;ic;Impartial Culture;o;process_id\n")




import numpy as np


def prepare_parties(culture_id=None, params=None):
    parties = []
    if culture_id == '2d_gaussian_party':
        for i in range(params['num_parties']):
            point = np.random.rand(1, 2)
            parties.append(point)
    elif culture_id in ['1d_gaussian_party', 'conitzer_party', 'walsh_party']:
        for i in range(params['num_parties']):
            point = np.random.rand(1, 1)
            parties.append(point)
    return parties


#!/usr/bin/env python

import os
import csv
import ast
import sys
import numpy as np
from tqdm import tqdm

try:
    from dotenv import load_dotenv
    load_dotenv()
    sys.path.append(os.environ["PATH"])
    from abcvoting.preferences import Profile
    from abcvoting import abcrules
except ImportError:
    Profile = None
    abcrules = None


def compute_abcvoting_rule(experiment=None, rule_name=None, committee_size=1, resolute=False):
    all_winning_committees = {}
    for election in tqdm(experiment.instances.values()):
        profile = Profile(election.num_candidates)
        if experiment.instance_type == 'ordinal':
            profile.add_voters(election.approval_votes)
        elif experiment.instance_type == 'approval':
            profile.add_voters(election.votes)
        try:
            winning_committees = abcrules.compute(rule_name, profile, committee_size,
                                                  algorithm="gurobi", resolute=resolute)
        except Exception:
            try:
                winning_committees = abcrules.compute(rule_name, profile, committee_size,
                                                      resolute=resolute)
            except:
                winning_committees = {}

        clean_winning_committees = []
        for committee in winning_committees:
            clean_winning_committees.append(set(committee))
        all_winning_committees[election.election_id] = clean_winning_committees
    store_committees_to_file(experiment.experiment_id, rule_name, all_winning_committees)


def store_committees_to_file(experiment_id, rule_name, all_winning_committees):
    path = os.path.join(os.getcwd(), "experiments", experiment_id, 'features',
                        f'{rule_name}.csv')
    with open(path, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(["election_id", "committee"])
        for election_id in all_winning_committees:
            writer.writerow([election_id, all_winning_committees[election_id]])


def import_committees_from_file(experiment_id, rule_name):
    all_winning_committees = {}
    path = os.path.join(os.getcwd(), "experiments", experiment_id, 'features',
                        f'{rule_name}.csv')
    with open(path, 'r', newline='') as csv_file:
        header = [h.strip() for h in csv_file.readline().split(';')]
        reader = csv.DictReader(csv_file, fieldnames=header, delimiter=';')
        for row in reader:
            try:
                winning_committees = ast.literal_eval(row['committee'])
            except:
                winning_committees = [set(row['committee'][2:-2].split(', '))]
            if len(winning_committees) == 0:
                winning_committees = [set()]
            all_winning_committees[row['election_id']] = winning_committees
    return all_winning_committees


def compute_not_abcvoting_rule(experiment=None, rule_name=None, committee_size=1, resolute=False):
    all_winning_committees = {}
    for election in tqdm(experiment.instances.values()):

        if rule_name == 'borda_c4':
            winning_committees = compute_borda_c4_rule(election, committee_size)
        elif rule_name == 'random':
            winning_committees = compute_random_rule(election, committee_size)

        all_winning_committees[election.election_id] = winning_committees
    store_committees_to_file(experiment.experiment_id, rule_name, all_winning_committees)


def compute_borda_c4_rule(election, committee_size=1):

    scores = np.zeros(election.num_candidates)
    for i, vote in enumerate(election.votes):
        max_score = len(election.approval_votes)
        for pos in range(max_score):
            scores[vote[pos]] += max_score - pos

    ranking = range(election.num_candidates)
    ranking = [x for _, x in sorted(zip(scores, ranking), reverse=True)]

    return [set(ranking[0:committee_size])]


def compute_random_rule(election, committee_size=1):

    candidates = np.random.permutation(election.num_candidates)
    return [set(candidates[0:committee_size])]



import numpy as np

from prefsampling.ordinal import impartial as generate_ordinal_ic_votes


def generate_real_identity_votes(num_voters=None, num_candidates=None):
    """ Generate real election that approximates identity (ID) """
    return [[j for j in range(num_candidates)] for _ in range(num_voters)]


def generate_real_uniformity_votes(num_voters=None, num_candidates=None):
    """ Generate real election that approximates uniformity (UN) """
    return generate_ordinal_ic_votes(num_voters=num_voters, num_candidates=num_candidates)


def generate_real_antagonism_votes(num_voters=None, num_candidates=None):
    """ Generate real election that approximates antagonism (AN) """
    return [[j for j in range(num_candidates)] for _ in range(int(num_voters / 2))] + \
           [[num_candidates - j - 1 for j in range(num_candidates)] for _ in
            range(int(num_voters / 2))]


def generate_real_stratification_votes(num_voters=None, num_candidates=None):
    """ Generate real election that approximates stratification (ST) """
    return [list(np.random.permutation(int(num_candidates/2))) +
             list(np.random.permutation([j for j in range(int(num_candidates/2), num_candidates)]))
            for _ in range(num_voters)]


# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 22.10.2021 #
# # # # # # # # # # # # # # # #



import logging
import numpy as np
try:
  from permanent import permanent
except:
  logging.warning("The 'permanent' module is not installed. Sampling matrices "
  "unavailable.")
import random
import math

import mapel.core.utils as utils

def _input_standarization(matrix):
  try: 
    if type(matrix) == list:
      matrix = np.asarray(matrix, dtype = 'int')
  except Exception as e:
    raise ValueError("Matrix provided for sampling cannot be made a numpy array", e)
  if matrix.shape[0] != matrix.shape[1]:
    raise ValueError("Matrix provided for sampling is not square", e)
  rows_sums = np.sum(matrix, axis = 0)
  cols_sums = np.sum(matrix, axis = 1)
  if not np.array_equal(rows_sums, cols_sums):
    raise ValueError("Matrix provided for sampling is not describing a valid \
    election---the row/column sums are not equal", e)
  return matrix

def _draw_vote(matrix):
  vote_matchings = []
  binary_matrix = np.sign(matrix)
  positions = np.arange(binary_matrix.shape[0])
  positions = positions[np.newaxis,:]
  candidates = np.arange(binary_matrix.shape[0] + 1)
  candidates = candidates[:, np.newaxis]
  binary_matrix = np.append(binary_matrix, positions, axis = 0)
  binary_matrix = np.hstack((binary_matrix, candidates))
  while len(vote_matchings) < len(matrix):
    taken_edge = None
    bin_matrix_view = binary_matrix[:-1,:-1]
    ground_perm = permanent.permanent(bin_matrix_view.astype('complex'))
    cands, poss = bin_matrix_view.nonzero()
    taken_edge = None
    while not taken_edge:
      if bin_matrix_view.shape == (1,1):
        taken_edge = (0,0)
        continue
      random_edge = random.choice(list(zip(cands, poss)))
      bin_matr_indices = range(bin_matrix_view.shape[0])
      cands_to_remain = [i for i in  bin_matr_indices if i!= random_edge[0]]
      pos_to_remain = [i for i in  bin_matr_indices if i!= random_edge[1]]
      without_taken_edge = bin_matrix_view[np.ix_(cands_to_remain,
      pos_to_remain)]
      took_edge_perm = permanent.permanent(without_taken_edge.astype('complex'))
      ratio = float(took_edge_perm.real)/float(ground_perm.real) 
      if math.isclose(ratio, 1.0):
        taken_edge = random_edge  
        continue 
      rand_num = random.random()
      if math.isclose(ratio, rand_num) or rand_num < ratio:
        taken_edge = random_edge
    orig_edge_cand = int(binary_matrix[taken_edge[0]][-1])
    orig_edge_pos = int(binary_matrix[-1][taken_edge[1]])
    vote_matchings.append((orig_edge_cand, orig_edge_pos))
    binary_matrix = np.delete(binary_matrix, taken_edge[0], axis = 0)
    binary_matrix = np.delete(binary_matrix, taken_edge[1], axis = 1)
  return vote_matchings

def sample_election_using_permanent(matrix):
  """
      Samples elections from a given position as follows:
      0. Interpret a given matrix as weighted bipartite graph (weights are
      entries)
      1. Sample uniformly at random a possible vote by sampling a perfect
      matching (that is unweighted)
      2. Subtract the generated vote from the original matrix and repeat step 0
      until the original matrix has only zero entries 

      Arguments:
        matrix: a position matrix either a list of lists or a numpy array

      Returns:
        The list of lists representing the votes realizing the given matrix
  """
  if not utils.is_module_loaded("permanent"):
    raise RuntimeError("Module 'permanent' was not loaded. Sampling elections "
    "from matrix impossible.")
  matrix = _input_standarization(matrix)
  votes_count = np.sum(matrix[0])
  curr_matrix = matrix.copy()
  votes = []
  while len(votes) < votes_count:
    matching = _draw_vote(curr_matrix)
    vote = [0] * len(matching)
    for cand, pos in matching:
      curr_matrix[cand][pos] -= 1
      vote[pos] = cand
    votes.append(vote)
  return votes



import numpy as np

from prefsampling.ordinal import impartial as generate_ordinal_ic_votes
from prefsampling.ordinal import urn as generate_urn_votes


def get_alliances(num_candidates, num_alliances):
    while True:
        alliances = np.random.choice([i for i in range(num_alliances)],
                                     size=num_candidates, replace=True)
        if len(set(alliances)) > 1:
            return alliances


def generate_ordinal_alliance_ic_votes(num_voters: int = None,
                                       num_candidates: int = None,
                                       params: dict = None):
    """ Return: ordinal votes from Impartial Culture with alliances """
    votes = generate_ordinal_ic_votes(num_voters, num_candidates)

    alliances = get_alliances(num_candidates, params['num_alliances'])

    return np.array(votes), alliances


def generate_ordinal_alliance_urn_votes(num_voters: int = None,
                                        num_candidates: int = None,
                                        params: dict = None):
    votes = generate_urn_votes(num_voters, num_candidates, params)

    alliances = get_alliances(num_candidates, params['num_alliances'])

    return np.array(votes), alliances


def generate_ordinal_alliance_norm_mallows_votes(num_voters: int = None,
                                                 num_candidates: int = None,
                                                 params: dict = None):

    params['phi'] = phi_from_normphi(num_candidates, params['normphi'])
    votes = generate_mallows_votes(num_voters, num_candidates, **params)

    alliances = get_alliances(num_candidates, params['num_alliances'])

    return np.array(votes), alliances


def generate_ordinal_alliance_euclidean_votes(num_voters: int = None,
                                              num_candidates: int = None,
                                              params: dict = None):
    dim = params['dim']
    num_alliances = params['num_alliances']

    voters = np.zeros([num_voters, dim])
    candidates = np.zeros([num_candidates, dim])
    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    voters = np.random.rand(num_voters, dim)
    centers = np.random.rand(num_alliances, dim)
    alliances = get_alliances(num_candidates, params['num_alliances'])

    candidates = []
    for c in range(num_candidates):
        candidates.append(np.random.normal(loc=centers[alliances[c]], scale=0.15, size=(1, dim)))

    for v in range(num_voters):
        for c in range(num_candidates):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(voters[v] - candidates[c], ord=params['dim'])

        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return votes, alliances


def assign_to_closest_center(candidates, centers):
    """
    Assign each candidate to the closest center.

    Parameters:
    - candidates: (num_candidates, dim) array, coordinates for each candidate point.
    - centers: (num_alliances, dim) array, coordinates for each center point.

    Returns:
    - assignment: (num_candidates,) array, index of the closest center for each candidate.
    """
    distances = np.linalg.norm(candidates[:, None] - centers, axis=2)
    assignment = np.argmin(distances, axis=1)
    return assignment


def generate_ordinal_alliance_allied_euclidean_votes(num_voters: int = None,
                                       num_candidates: int = None,
                                       params: dict = None):
    dim = params['dim']
    num_alliances = params['num_alliances']

    votes = np.zeros([num_voters, num_candidates], dtype=int)
    distances = np.zeros([num_voters, num_candidates], dtype=float)

    while True:
        voters = np.random.rand(num_voters, dim)
        centers = np.random.rand(num_alliances, dim)
        candidates = np.random.rand(num_candidates, dim)

        alliances = assign_to_closest_center(candidates, centers)
        if len(set(alliances)) == num_alliances:
            break


    for v in range(num_voters):
        for c in range(num_candidates):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(voters[v] - candidates[c])

        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return votes, alliances



#!/usr/bin/env python

import numpy as np


def get_single_crossing_matrix(num_candidates: int) -> np.ndarray:
    return get_single_crossing_vectors(num_candidates).transpose()


def get_single_crossing_vectors(num_candidates: int) -> np.ndarray:
    matrix = np.zeros([num_candidates, num_candidates])

    for i in range(num_candidates):
        for j in range(num_candidates - i):
            matrix[i][j] = i + j

    for i in range(num_candidates):
        for j in range(i):
            matrix[i][j] += 1

    sums = [1]
    for i in range(num_candidates):
        sums.append(sums[i] + i)

    for i in range(num_candidates):
        matrix[i][i] += sums[i]
        matrix[i][num_candidates - i - 1] -= i

    for i in range(num_candidates):
        denominator = sum(matrix[i])
        for j in range(num_candidates):
            matrix[i][j] /= denominator

    return matrix



#!/usr/bin/env python

import copy
import logging

from mapel.core.objects.Family import Family
from mapel.elections.objects.OrdinalElection import OrdinalElection
from mapel.elections.objects.ApprovalElection import ApprovalElection
from mapel.core.utils import *
from mapel.elections.cultures.params import *


class ElectionFamily(Family):
    """ Family of elections: a set of elections from the same election culture_id """

    def __init__(self,
                 culture_id: str = None,
                 family_id: str = 'none',
                 params: dict = None,
                 size: int = 1,
                 label: str = None,
                 color: str = "black",
                 alpha: float = 1.,
                 ms: int = 20,
                 show: bool = True,
                 marker: str = 'o',
                 starting_from: int = 0,
                 path: dict = None,
                 single: bool = False,

                 instance_type = None,
                 num_candidates: int = None,
                 num_voters: int = None,
                 election_ids=None,
                 **kwargs):

        super().__init__(culture_id=culture_id,
                         family_id=family_id,
                         params=params,
                         size=size,
                         label=label,
                         color=color,
                         alpha=alpha,
                         ms=ms,
                         show=show,
                         marker=marker,
                         starting_from=starting_from,
                         path=path,
                         single=single,
                         instance_ids=election_ids,
                         **kwargs)

        self.num_candidates = num_candidates
        self.num_voters = num_voters
        self.instance_type = instance_type


    def __getattr__(self, attr):
        if attr == 'election_ids':
            return self.instance_ids
        else:
            return self.__dict__[attr]

    def __setattr__(self, name, value):
        if name == "election_ids":
            return setattr(self, 'instance_ids', value)
        else:
            self.__dict__[name] = value

    def prepare_family(self,
                       experiment_id=None,
                       is_exported=True,
                       store_points=False,
                       is_aggregated=True,
                       instance_type=None):

        if instance_type is not None:
            self.instance_type = instance_type

        if self.instance_type == 'ordinal':

            elections = {}
            _keys = []
            for j in range(self.size):

                params = copy.deepcopy(self.params)

                variable = None
                path = self.path
                if path is not None and 'variable' in path:
                    new_params, variable = get_params_for_paths(self, j)
                    if params is None:
                        params = {}
                    params = {**params, **new_params}

                election_id = get_instance_id(self.single, self.family_id, j)
                election = OrdinalElection(experiment_id, election_id,
                                           culture_id=self.culture_id,
                                           num_voters=self.num_voters,
                                           label=self.label,
                                           num_candidates=self.num_candidates,
                                           variable=variable,
                                           is_imported=False,
                                           **params
                                           )

                election.prepare_instance(is_exported=is_exported, is_aggregated=is_aggregated)

                if store_points:
                    try:
                        election.points['voters'] = election.import_ideal_points('voters')
                        election.points['candidates'] = election.import_ideal_points('candidates')
                    except:
                        pass

                election.compute_potes()

                elections[election_id] = election

                _keys.append(election_id)

            self.election_ids = _keys

        elif self.instance_type == 'approval':

            elections = {}
            _keys = []
            for j in range(self.size):

                params = copy.deepcopy(self.params)

                variable = None
                path = self.path
                if path is not None and 'variable' in path:
                    new_params, variable = get_params_for_paths(self, j)
                    params = {**params, **new_params}

                if self.culture_id in {'all_votes'}:
                    params['iter_id'] = j

                if self.culture_id in {'crate'}:
                    new_params = get_params_for_crate(j)
                    params = {**params, **new_params}

                election_id = get_instance_id(self.single, self.family_id, j)

                election = ApprovalElection(experiment_id,
                                            election_id,
                                            culture_id=self.culture_id,
                                            num_voters=self.num_voters,
                                            label=self.label,
                                            num_candidates=self.num_candidates,
                                            ballot_type=self.instance_type,
                                            variable=variable,
                                            is_imported=False,
                                            **params
                                            )
                election.prepare_instance(is_exported=is_exported, is_aggregated=is_aggregated)

                election.votes_to_approvalwise_vector()

                elections[election_id] = election

                _keys.append(election_id)

            self.election_ids = _keys

        else:
            logging.warning('No such instance type!')
            return None

        return elections

    def add_election(self, election):
        self.size += 1
        self.election_ids.append(election.instance_id)



# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 12.10.2021 #
# # # # # # # # # # # # # # # #



#!/usr/bin/env python

import os
from contextlib import suppress
import scipy.special

try:
    import cplex
except ImportError:
    cplex = None


# SPEARMAN
def generate_ilp_spearman_distance(lp_file_name, votes_1, votes_2, params):
    lp_file = open(lp_file_name, 'w')
    lp_file.write("Minimize\n")

    first = True
    for k in range(params['voters']):
        for l in range(params['voters']):

            vote_1 = votes_1[k]
            vote_2 = votes_2[l]

            pote_1 = [0] * params['candidates']
            pote_2 = [0] * params['candidates']

            for i in range(params['candidates']):
                pote_1[vote_1[i]] = i
                pote_2[vote_2[i]] = i

            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False

                    weight = abs(pote_1[i] - pote_2[j])

                    lp_file.write(
                        str(weight) + " P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(
                            j))
    lp_file.write("\n")

    lp_file.write("Subject To\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "M" + "i" + str(i) + "j" + str(j) + " <= 0" + "\n")

                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "N" + "k" + str(k) + "l" + str(l) + " <= 0" + "\n")

    for k in range(params['voters']):
        first = True
        for l in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    for l in range(params['voters']):
        first = True
        for k in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    for i in range(params['candidates']):
        first = True
        for j in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    for j in range(params['candidates']):
        first = True
        for i in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    for k in range(params['voters']):
        for i in range(params['candidates']):
            first = True
            for l in range(params['voters']):
                for j in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    for l in range(params['voters']):
        for j in range(params['candidates']):
            first = True
            for k in range(params['voters']):
                for i in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    lp_file.write("Binary\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write(
                        "P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j) + "\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            lp_file.write("N" + "k" + str(k) + "l" + str(l) + "\n")

    for i in range(params['candidates']):
        for j in range(params['candidates']):
            lp_file.write("M" + "i" + str(i) + "j" + str(j) + "\n")

    lp_file.write("End\n")


def solve_ilp_distance(path):
    cp_lp = cplex.Cplex(path)
    cp_lp.set_results_stream(None)

    cp_lp.parameters.threads.set(1)

    try:
        cp_lp.solve()

    except cplex.CplexSolverError:
        print("Exception raised during solve")
        return

    return cp_lp.solution.get_objective_value()


def spearman_cost(single_votes_1, single_votes_2, params, perm):
    pote_1 = [0] * params['candidates']
    pote_2 = [0] * params['candidates']

    for i in range(params['candidates']):
        id_1 = int(perm[0][single_votes_1[i]])
        pote_1[id_1] = i
        id_2 = int(perm[1][single_votes_2[i]])
        pote_2[id_2] = i

    total_diff = 0.
    for i in range(params['candidates']):
        local_diff = float(abs(pote_1[i] - pote_2[i]))
        total_diff += local_diff

    return total_diff


def spearman_cost_per_cand(single_votes_1, single_votes_2, params, perm):
    pote_1 = [0] * params['candidates']
    pote_2 = [0] * params['candidates']

    for i in range(params['candidates']):
        id_1 = int(perm[0][single_votes_1[i]])
        pote_1[id_1] = i
        id_2 = int(perm[1][single_votes_2[i]])
        pote_2[id_2] = i

    cand_diff = [0] * params['candidates']
    for i in range(params['candidates']):
        cand_diff[i] = float(abs(pote_1[i] - pote_2[i]))

    return cand_diff


def remove_lp_file(path):
    """ Safely remove lp file """
    with suppress(OSError):
        os.remove(path)


# SWAP
def generate_ilp_swap_distance(lp_file_name, votes_1, votes_2, params):

    all = int(params['voters'] * scipy.special.binom(params['candidates'], 2))

    lp_file = open(lp_file_name, 'w')
    lp_file.write("Minimize\n")

    first = True
    for k in range(params['voters']):
        for l in range(params['voters']):

            # pote_1 = [0] * params['candidates']
            # pote_2 = [0] * params['candidates']
            #
            # for i in range(params['candidates']):
            #     pote_1[votes_1[k][i]] = i
            #     pote_2[votes_2[l][i]] = i

            vote_1 = votes_1[k]
            vote_2 = votes_2[l]

            pote_1 = [list(vote_1).index(i) for i, _ in enumerate(vote_1)]
            pote_2 = [list(vote_2).index(i) for i, _ in enumerate(vote_2)]

            # pote_1 = [0] * params['candidates']
            # pote_2 = [0] * params['candidates']

            # for i in range(params['candidates']):
            #     pote_1[vote_1[i]] = i
            #     pote_2[vote_2[i]] = i

            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(i1+1, params['candidates']):
                    # for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):

                            if i1 == i2 or j1 == j2:
                                continue

                            # if k==1 and l==0 and i1==0 and j1==2 and i2==1 and j2==1:
                            #     print(pote_1[i1], pote_2[j1], pote_1[i2], pote_2[j2])

                            # if (pote_1[i1] > pote_2[j1] and pote_1[i2] < pote_2[j2]) or \
                            #         (pote_1[i1] < pote_2[j1] and pote_1[i2] > pote_2[j2]):
                            if (pote_1[i1] > pote_1[i2] and pote_2[j1] < pote_2[j2]) or \
                                    (pote_1[i1] < pote_1[i2] and pote_2[j1] > pote_2[j2]) :

                                if not first:
                                    lp_file.write(" + ")
                                first = False
                                lp_file.write(" R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(
                                    j1) + "i" + str(i2) + "j" + str(j2))
    lp_file.write("\n")

    lp_file.write("Subject To\n")

    # N=1
    for k in range(params['voters']):
        first = True
        for l in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    # N=1
    for l in range(params['voters']):
        first = True
        for k in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    # M=1
    for i in range(params['candidates']):
        first = True
        for j in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    # M=1
    for j in range(params['candidates']):
        first = True
        for i in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    # P=1
    for k in range(params['voters']):
        for i in range(params['candidates']):
            first = True
            for l in range(params['voters']):
                for j in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    # P=1
    for l in range(params['voters']):
        for j in range(params['candidates']):
            first = True
            for k in range(params['voters']):
                for i in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    # P<N, P<M
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "N" + "k" + str(k) + "l" + str(l) + " <= 0" + "\n")

                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "M" + "i" + str(i) + "j" + str(j) + " <= 0" + "\n")

    # All
    first = True
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(i1+1, params['candidates']):
                    # for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):

                            if i1 == i2 or j1 == j2:
                                continue

                            if not first:
                                lp_file.write(" + ")
                            first = False
                            lp_file.write("R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(j1) + "i" + str(
                                i2) + "j" + str(j2))
    lp_file.write(" = " + str(all) + "\n")

    # R<P, R<P
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(i1+1, params['candidates']):
                    # for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):

                            if i1 == i2 or j1 == j2:
                                continue

                            lp_file.write("R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(j1) + "i" + str(
                                i2) + "j" + str(j2))
                            lp_file.write(" - " + "P" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(
                                j1) + " <= 0" + "\n")

                            lp_file.write("R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(j1) + "i" + str(
                                i2) + "j" + str(j2))
                            lp_file.write(" - " + "P" + "k" + str(k) + "l" + str(l) + "i" + str(i2) + "j" + str(
                                j2) + " <= 0" + "\n")

    lp_file.write("Binary\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(i1+1, params['candidates']):
                    # for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):

                            if i1 == i2 or j1 == j2:
                                continue

                            lp_file.write("R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(j1) + "i" + str(
                                i2) + "j" + str(j2) + "\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j) + "\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            lp_file.write("N" + "k" + str(k) + "l" + str(l) + "\n")

    for i in range(params['candidates']):
        for j in range(params['candidates']):
            lp_file.write("M" + "i" + str(i) + "j" + str(j) + "\n")

    lp_file.write("End\n")


def generate_ilp_distance_swap_short(lp_file_name, votes_1, votes_2, params):
    lp_file = open(lp_file_name, 'w')
    lp_file.write("Minimize\n")  # obj: ")

    first = True
    for k in range(params['voters']):
        for l in range(params['voters']):

            pote_1 = [0] * params['candidates']
            pote_2 = [0] * params['candidates']

            for i in range(params['candidates']):
                pote_1[votes_1[k][i]] = i
                pote_2[votes_2[l][i]] = i

            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):



                            if not first:
                                lp_file.write(" + ")
                            first = False

                            if pote_1[i1] > pote_2[i2] and pote_1[j1] < pote_2[j2]:
                                weight = 1
                            elif pote_1[i1] < pote_2[i2] and pote_1[j1] > pote_2[j2]:
                                weight = 1
                            else:
                                weight = 0

                            lp_file.write(str(weight) + " R" + str(k) + str(l) + str(i1) + str(j1) + str(i2) + str(j2))
    lp_file.write("\n")

    lp_file.write("Subject To\n")

    # N=1
    for k in range(params['voters']):
        first = True
        for l in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    # N=1
    for l in range(params['voters']):
        first = True
        for k in range(params['voters']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("N" + "k" + str(k) + "l" + str(l))
        lp_file.write(" = 1" + "\n")

    # M=1
    for i in range(params['candidates']):
        first = True
        for j in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    # M=1
    for j in range(params['candidates']):
        first = True
        for i in range(params['candidates']):
            if not first:
                lp_file.write(" + ")
            first = False
            lp_file.write("M" + "i" + str(i) + "j" + str(j))
        lp_file.write(" = 1" + "\n")

    # P=1
    for k in range(params['voters']):
        for i in range(params['candidates']):
            first = True
            for l in range(params['voters']):
                for j in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    # P=1
    for l in range(params['voters']):
        for j in range(params['candidates']):
            first = True
            for k in range(params['voters']):
                for i in range(params['candidates']):
                    if not first:
                        lp_file.write(" + ")
                    first = False
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
            lp_file.write(" = 1" + "\n")

    # P<N, P<M
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "N" + "k" + str(k) + "l" + str(l) + " <= 0" + "\n")

                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j))
                    lp_file.write(" - " + "M" + "i" + str(i) + "j" + str(j) + " <= 0" + "\n")

    # R=M
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    first = True
                    for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):
                            if not first:
                                lp_file.write(" + ")
                            first = False
                            lp_file.write("R" + "k" + str(k) + "l" + str(l) + str(i1) + str(j1) + str(i2) + str(j2))
                    lp_file.write(" = " + str(params['candidates']) + "\n")

    # R=M
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i2 in range(params['candidates']):
                for j2 in range(params['candidates']):
                    first = True
                    for i1 in range(params['candidates']):
                        for j1 in range(params['candidates']):
                            if not first:
                                lp_file.write(" + ")
                            first = False
                            lp_file.write("R" + str(k) + str(l) + str(i1) + str(j1) + str(i2) + str(j2))
                    lp_file.write(" = " + str(params['candidates']) + "\n")

    # R<P, R<P
    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):

                            lp_file.write("R" + str(k) + str(l) + str(i1) + str(j1) + str(i2) + str(j2))
                            lp_file.write(" - " + "P" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(
                                j1) + " <= 0" + "\n")


                            lp_file.write("R" + str(k) + str(l) + str(i1) + str(j1) + str(i2) + str(j2))
                            lp_file.write(" - " + "P" + "k" + str(k) + "l" + str(l) + "i" + str(i2) + "j" + str(
                                j2) + " <= 0" + "\n")

    lp_file.write("Binary\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i1 in range(params['candidates']):
                for j1 in range(params['candidates']):
                    for i2 in range(params['candidates']):
                        for j2 in range(params['candidates']):
                            lp_file.write("R" + str(k) + str(l) + str(i1) + str(j1) + str(i2) + str(j2) + "\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            for i in range(params['candidates']):
                for j in range(params['candidates']):
                    lp_file.write("P" + "k" + str(k) + "l" + str(l) + "i" + str(i) + "j" + str(j) + "\n")

    for k in range(params['voters']):
        for l in range(params['voters']):
            lp_file.write("N" + "k" + str(k) + "l" + str(l) + "\n")

    for i in range(params['candidates']):
        for j in range(params['candidates']):
            lp_file.write("M" + "i" + str(i) + "j" + str(j) + "\n")

    lp_file.write("End\n")


def solve_ilp_distance_swap(path, votes_1, votes_2, params):

    cp_lp = cplex.Cplex(path)
    cp_lp.set_results_stream(None)

    cp_lp.parameters.threads.set(1)

    try:
        cp_lp.solve()

    except:
        print("Exception raised during solve")
        return

    # total = 0
    # for k in range(params['voters']):
    #     for l in range(params['voters']):
    #
    #         pote_1 = [0] * params['candidates']
    #         pote_2 = [0] * params['candidates']
    #
    #         for i in range(params['candidates']):
    #             pote_1[votes_1[k][i]] = i
    #             pote_2[votes_2[l][i]] = i
    #
    #         for i1 in range(params['candidates']):
    #             for j1 in range(params['candidates']):
    #                 for i2 in range(i1+1, params['candidates']):
    #                     for j2 in range(params['candidates']):
    #
    #                         if i1 == i2 or j1 == j2:
    #                             continue
    #
    #                         if pote_1[i1] > pote_2[j1] and pote_1[i2] < pote_2[j2]:
    #                             weight = 1
    #                         elif pote_1[i1] < pote_2[j1] and pote_1[i2] > pote_2[j2]:
    #                             weight = 1
    #                         else:
    #                             weight = 0
    #
    #                         A = "R" + "k" + str(k) + "l" + str(l) + "i" + str(i1) + "j" + str(j1) + "i" + str(i2) + "j" + str(j2)
    #                         # print(weight, int(cp_lp.solution.get_values(A)))
    #                         if int(cp_lp.solution.get_values(A)) == 1:
    #                             print(A, weight)
    #                         total += weight * int(cp_lp.solution.get_values(A))
    # print(f'total {total}')

    return cp_lp.solution.get_objective_value()


import csv
import os
import numpy as np
import random

import mapel.elections.persistence.election_exports as exports


def convert_pb_to_app(experiment, num_candidates=100, num_voters=100, model='pabulib',
                      aggregated=True, num_instances=1):

    main_path = os.path.join(os.getcwd(), "election", experiment.experiment_id, "source")

    paths = []
    for name in os.listdir(main_path):
        # print(name)
        paths.append(f'election/{experiment.experiment_id}/source/{name}')

    for p in range(num_instances):

        random_index = random.randint(0, len(paths) - 1)
        path = paths[random_index]

        print(p, path)

        votes = {}
        projects = {}

        # import from .pb file
        with open(path, 'r', newline='', encoding="utf-8") as csvfile:
            meta = {}
            section = ""
            header = []
            reader = csv.reader(csvfile, delimiter=';')
            for row in reader:
                if str(row[0]).strip().lower() in ["meta", "projects", "votes"]:
                    section = str(row[0]).strip().lower()
                    header = next(reader)
                elif section == "meta":
                    meta[row[0]] = row[1].strip()
                elif section == "projects":
                    projects[row[0]] = {}
                    for it, key in enumerate(header[1:]):
                        projects[row[0]][key.strip()] = row[it+1].strip()
                elif section == "votes":
                    votes[row[0]] = {}
                    for it, key in enumerate(header[1:]):
                        votes[row[0]][key.strip()] = row[it+1].strip()

        # extract approval ballots from votes
        approval_votes = []
        for vote in votes.values():
            vote = vote['vote'].split(',')
            approval_votes.append(set(vote))

        # randomly map projects to [0,1,2...,num_projects]
        mapping = {}
        perm = np.random.permutation(len(projects))
        for i, key in enumerate(projects.keys()):
            mapping[key] = perm[i]

        approval_votes = [[mapping[i] for i in vote] for vote in approval_votes]

        # cut projects to [0,1,2...,k]
        approval_votes_cut = []
        for vote in approval_votes:
            vote_cut = set()
            for c in vote:
                if c < num_candidates:
                    vote_cut.add(c)
            approval_votes_cut.append(vote_cut)

        # cut the voters:
        perm = np.random.permutation(len(votes))
        # print(len(perm))
        final_approval_votes_cut = []

        ctr = 0
        for i in range(len(approval_votes_cut)):
            j = perm[i]
            vote = approval_votes_cut[j]
            if vote != set():
                final_approval_votes_cut.append(vote)
                ctr += 1

            if ctr == num_voters:
                break


        # is_exported in .app file
        name = name.replace('.pb', '')
        path = f'election/{experiment.experiment_id}/elections/{model}_{p}.app'
        # num_candidates = num_candidates
        # num_voters = len(votes)
        params = {}
        ballot = 'approval'

        election = None
        model_id = model
        exports.export_votes_to_file(election, model_id, num_candidates, num_voters,
                                     params, path, ballot, votes=final_approval_votes_cut,
                                     is_aggregated=aggregated)




#!/usr/bin/env python

import copy
import csv
import itertools
import logging
import math
import os
from abc import abstractmethod

import matplotlib.pyplot as plt
import numpy as np
import numpy.linalg as la
from sklearn.manifold import MDS

import mapel.elections.persistence.election_exports as exports
import mapel.elections.persistence.election_imports as imports
from mapel.core.glossary import *
from mapel.core.inner_distances import l2
from mapel.core.objects.Instance import Instance
from mapel.elections.features_ import get_local_feature
from mapel.elections.other.winners import compute_sntv_winners, compute_borda_winners, \
    compute_stv_winners

OBJECT_TYPES = ['vote', 'candidate']


class Election(Instance):

    def __init__(self,
                 experiment_id=None,
                 election_id=None,
                 culture_id=None,
                 votes=None,
                 ballot_type: str = 'ordinal',
                 num_voters: int = None,
                 num_candidates: int = None,
                 label=None,
                 fast_import=False,
                 is_shifted=False,
                 is_imported=False,
                 is_exported=True,
                 **kwargs):

        super().__init__(experiment_id=experiment_id,
                         instance_id=election_id,
                         culture_id=culture_id,
                         **kwargs)

        self.election_id = election_id
        self.ballot_type = ballot_type
        self.label = label
        self.num_voters = num_voters
        self.num_candidates = num_candidates
        self.votes = votes
        self.is_exported = is_exported
        self.winners = None
        self.alternative_winners = {}
        self.fake = culture_id in LIST_OF_FAKE_MODELS
        self.potes = None
        self.features = {}
        self.object_type = 'vote'
        self.points = {}
        self.is_shifted = is_shifted
        self.is_imported = is_imported
        self.fast_import = fast_import

        self.import_distances()
        self.import_coordinates()

    def import_distances(self):
        """
        Import distances from a .csv file.

        :return: None
        """
        self.distances = {}
        if not self.fast_import:
            for object_type in OBJECT_TYPES:
                try:
                    self.distances[object_type] = imports.import_distances(self, object_type)
                except:
                    pass

    def import_coordinates(self):
        """
        Import coordinates from a .csv file.

        :return: None
        """
        self.coordinates = {}
        for object_type in OBJECT_TYPES:
            try:
                self.coordinates[object_type] = imports.import_coordinates(self, object_type)
            except:
                pass

    def get_distances(self, object_type):
        try:
            return self.distances[object_type]
        except:
            self.distances[object_type] = imports.import_distances(self, object_type)
            return self.distances[object_type]

    def get_coordiantes(self, object_type):
        try:
            return self.coordinates[object_type]
        except:
            self.coordinates[object_type] = imports.import_coordinates(self, object_type)
            return self.coordinates[object_type]

    def set_default_object_type(self, object_type):
        self.object_type = object_type

    def import_matrix(self) -> np.ndarray:
        file_name = f'{self.election_id}.csv'
        path = os.path.join(os.getcwd(), "experiments", self.experiment_id, 'matrices', file_name)
        matrix = np.zeros([self.num_candidates, self.num_candidates])

        with open(path, 'r', newline='') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=';')
            for i, row in enumerate(reader):
                for j, candidate_id in enumerate(row):
                    matrix[i][j] = row[candidate_id]
        return matrix

    def compute_potes(self, mapping=None):
        """ Convert votes to positional votes (called potes) """
        if not self.fake:
            if mapping is None:
                self.potes = np.array([[list(vote).index(i) for i, _ in enumerate(vote)]
                                       for vote in self.votes])
            else:
                self.potes = np.array([[list(vote).index(mapping[i]) for i, _ in enumerate(vote)]
                                       for vote in self.votes])
            return self.potes

    def vector_to_interval(self, vector, precision=None) -> list:
        # discreet version for now
        interval = []
        w = int(precision / self.num_candidates)
        for i in range(self.num_candidates):
            for j in range(w):
                interval.append(vector[i] / w)
        return interval

    def compute_alternative_winners(self, method=None, party_id=None, num_winners=None):

        election_without_party_id = remove_candidate_from_election(copy.deepcopy(self),
                                                                   party_id, num_winners)
        election_without_party_id = map_the_votes(election_without_party_id, party_id, num_winners)

        if method == 'sntv':
            winners_without_party_id = compute_sntv_winners(election=election_without_party_id,
                                                            num_winners=num_winners)
        elif method == 'borda':
            winners_without_party_id = compute_borda_winners(election=election_without_party_id,
                                                             num_winners=num_winners)
        elif method == 'stv':
            winners_without_party_id = compute_stv_winners(election=election_without_party_id,
                                                           num_winners=num_winners)
        else:
            winners_without_party_id = []

        self.alternative_winners[party_id] = unmap_the_winners(winners_without_party_id, party_id,
                                                               num_winners)

    def print_euclidean_voters_and_candidates_map(self, show=True, radius=None, name=None,
                                                  alpha=0.5, s=30, circles=False,
                                                  saveas=None,
                                                  object_type=None, double_gradient=False):

        plt.figure(figsize=(6.4, 6.4))

        X_voters = []
        Y_voters = []
        for i in range(self.num_voters):
            x = self.points['voters'][i][0]
            y = self.points['voters'][i][1]
            plt.scatter(x, y, color=[0, y, x], s=s, alpha=0.3)
            # X_voters.append(election.points['voters'][i][0])
            # Y_voters.append(election.points['voters'][i][1])
        # plt.scatter(X_voters, Y_voters, color='grey', s=s, alpha=0.1)

        X_candidates = []
        Y_candidates = []
        for i in range(self.num_candidates):
            X_candidates.append(self.points['candidates'][i][0])
            Y_candidates.append(self.points['candidates'][i][1])
        plt.scatter(X_candidates, Y_candidates, color='red', s=s * 2, alpha=0.9)

        if radius:
            plt.xlim([-radius, radius])
            plt.ylim([-radius, radius])
        plt.title(self.label, size=38)
        plt.axis('off')

        if saveas is None:
            saveas = f'{self.label}_euc.png'

        file_name = os.path.join(os.getcwd(), "images", name, f'{saveas}.png')
        plt.savefig(file_name, bbox_inches='tight', dpi=100)
        if show:
            plt.show()
        else:
            plt.clf()


    @abstractmethod
    def compute_distances(self):
        pass

    #DIV-MERGE
    def embed(self, algorithm='MDS', object_type=None, virtual=False):

        if object_type is None:
            object_type = self.object_type

        MDS_object = MDS(n_components=2,
                         dissimilarity='precomputed',
                         normalized_stress='auto',
                # max_iter=1000,
                # n_init=20,
                # eps=1e-6,
            )

        #DIV-MERGE
        if algorithm == 'PCA':
            self.coordinates[object_type] = pca(self.distances[object_type])
        else:
            self.coordinates[object_type] = MDS_object.fit_transform(self.distances[object_type])

        if object_type == 'vote':
            length = self.num_options
        elif object_type == 'candidate':
            # length = experiment_id.num_candidates
            pass
        else:
            logging.warning('No such type of object!')
            length = None

        # ADJUST
        # find max dist
        # if (not ('identity' in election.culture_id.lower() and object_type=='vote')) \
        #         and (not ('approval_id' in election.culture_id.lower() and object_type=='vote')):
        if not self.all_dist_zeros(object_type):
            dist = np.zeros(
                [len(self.coordinates[object_type]), len(self.coordinates[object_type])])
            for pos_1, pos_2 in itertools.combinations(
                    [i for i in range(len(self.coordinates[object_type]))],
                    2):
                dist[pos_1][pos_2] = l2(self.coordinates[object_type][pos_1],
                                        self.coordinates[object_type][pos_2])

            result = np.where(dist == np.amax(dist))
            id_1 = result[0][0]
            id_2 = result[1][0]

            # rotate
            a = id_1
            b = id_2

            try:
                d_x = self.coordinates[object_type][a][0] - self.coordinates[object_type][b][0]
                d_y = self.coordinates[object_type][a][1] - self.coordinates[object_type][b][1]
                alpha = math.atan(d_x / d_y)
                self.rotate(alpha - math.pi / 2., object_type)
                self.rotate(math.pi / 4., object_type)
            except Exception:
                pass

            # PUT heavier corner in the left lower part
            if self.coordinates[object_type][a][0] < self.coordinates[object_type][b][0]:
                left = a
                right = b
            else:
                left = b
                right = a
            try:
                left_ctr = 0
                right_ctr = 0
                for v in range(length):
                    d_left = l2(self.coordinates[object_type][left],
                                self.coordinates[object_type][v])
                    d_right = l2(self.coordinates[object_type][right],
                                 self.coordinates[object_type][v])
                    if d_left < d_right:
                        left_ctr += 1
                    else:
                        right_ctr += 1

                if left_ctr < right_ctr:
                    self.rotate(math.pi, object_type)

            except Exception:
                pass


        if object_type == 'vote':
            length = self.num_options
        elif object_type == 'candidate':
            length = self.num_candidates
            pass

        if self.is_exported and not virtual:
            exports.export_coordinates(self, object_type=object_type, length=length)

    def all_dist_zeros(self, object_type):
        if np.abs(self.distances[object_type]).sum():
            return False
        else:
            return True

    @staticmethod
    def rotate_point(cx, cy, angle, px, py) -> (float, float):
        """ Rotate two-dimensional point by an angle """
        s, c = math.sin(angle), math.cos(angle)
        px -= cx
        py -= cy
        x_new, y_new = px * c - py * s, px * s + py * c
        px, py = x_new + cx, y_new + cy

        return px, py

    def rotate(self, angle, object_type) -> None:
        """ Rotate all the points by a given angle """
        for instance_id in range(len(self.coordinates[object_type])):
            self.coordinates[object_type][instance_id][0], \
            self.coordinates[object_type][instance_id][1] = \
                self.rotate_point(0.5, 0.5, angle, self.coordinates[object_type][instance_id][0],
                                  self.coordinates[object_type][instance_id][1])

    def compute_feature(self, feature_id, feature_long_id=None, **kwargs):
        if feature_long_id is None:
            feature_long_id = feature_id
        feature = get_local_feature(feature_id)
        self.features[feature_long_id] = feature(self, **kwargs)

    def get_feature(self,
                    feature_id,
                    feature_long_id=None,
                    overwrite=False,
                    **kwargs):
        if feature_long_id is None:
            feature_long_id = feature_id
        if feature_id not in self.features or overwrite:
            self.compute_feature(feature_id, feature_long_id, **kwargs)
        return self.features[feature_long_id]


def map_the_votes(election, party_id, party_size) -> Election:
    new_votes = [[] for _ in range(election.num_voters)]
    for i in range(election.num_voters):
        for j in range(election.num_candidates):
            if election.votes[i][j] >= party_id * party_size:
                new_votes[i].append(election.votes[i][j] - party_size)
            else:
                new_votes[i].append(election.votes[i][j])
    election.votes = new_votes
    return election


def unmap_the_winners(winners, party_id, party_size):
    new_winners = []
    for j in range(len(winners)):
        if winners[j] >= party_id * party_size:
            new_winners.append(winners[j] + party_size)
        else:
            new_winners.append(winners[j])
    return new_winners


def remove_candidate_from_election(election, party_id, party_size) -> Election:
    for vote in election.votes:
        for i in range(party_size):
            _id = party_id * party_size + i
            vote.remove(_id)
    election.num_candidates -= party_size
    return election

#DIV-MERGE
def pca(distance_matrix):
    # print(distance_matrix)
    # df = pd.read_csv("http://rosetta.reltech.org/TC/v15/Mapping/data/dist-Aus.csv")
    # A = df.values.T[1:].astype(float)
    A = distance_matrix
    # square it
    A = A ** 2
    # centering matrix
    n = A.shape[0]
    # J_c = 1. / n * (np.eye(n) - 1 + (n - 1) * np.eye(n))
    J_c = np.eye(n) - 1./n

    # perform double centering
    B = -0.5 * np.matmul(np.matmul(J_c, A), J_c)

    # find eigenvalues and eigenvectors
    eigen_val = la.eig(B)[0]
    eigen_vec = la.eig(B)[1].T

    eigen_vec_real = np.round(np.real(eigen_vec), 5)
    # eigen_vec_imag = np.round(np.imag(eigen_vec), 5)
    # if np.abs(eigen_vec_imag).sum() > 1:
    #     print("Complex eigenvectors!")
    #     print(np.abs(eigen_vec_imag).sum())
        # print(eigen_vec_imag)
        # print(eigen_vec)

    eigen_val_real = np.round(np.real(eigen_val), 5)
    # eigen_val_imag = np.round(np.imag(eigen_val), 5)
    # print(eigen_val_real)
    # print(eigen_val_imag)
    # if np.abs(eigen_val_imag).sum() > 1:
    #     print("Complex eigenvalues!")
        # print(eigen_val_imag)
        # print(eigen_val)
    bests = np.argsort(-eigen_val_real)
    i = bests[0]
    j = bests[1]

    PC1 = np.sqrt(eigen_val_real[i]) * eigen_vec_real[i]
    PC2 = np.sqrt(eigen_val_real[j]) * eigen_vec_real[j]
    res = np.array([[x, y] for x, y in zip(PC1, PC2)])
    # print(PC1)
    # print(PC2)
    return res






import numpy as np
import scipy.special
import math


def justified_ratio(election, feature_params) -> float:
    # 1-large, 1-cohesive
    election.compute_reverse_approvals()
    threshold = election.num_voters / feature_params['committee_size']
    covered = set()
    for _set in election.reverse_approvals:
        if len(_set) >= threshold:
            covered = covered.union(_set)
    print(len(covered) / float(election.num_voters))
    return len(covered) / float(election.num_voters)


def abstract(election) -> float:
    n = election.num_voters
    election.votes_to_approvalwise_vector()
    vector = election.approvalwise_vector
    total_value = 0
    for i in range(election.num_candidates):
        k = vector[i] * n
        x = scipy.special.binom(n, k)
        x = math.log(x)
        total_value += x
    return total_value


def borda_std(election):
    all_scores = np.zeros(election.num_candidates)
    vectors = election.get_vectors()
    for i in range(election.num_candidates):
        for j in range(election.num_candidates):
            all_scores[i] += vectors[i][j] * (election.num_candidates - j - 1) * election.num_voters
    return np.std(all_scores)


def get_effective_num_candidates(election, mode='Borda') -> float:
    """ Compute effective number of candidates """

    c = election.num_candidates
    vectors = election.votes_to_positionwise_matrix()

    if mode == 'Borda':
        all_scores = [sum([vectors[j][i] * (c - i - 1) for i in range(c)]) / (c * (c - 1) / 2)
                      for j in range(c)]
    elif mode == 'Plurality':
        all_scores = [sum([vectors[j][i] for i in range(1)]) for j in range(c)]
    else:
        all_scores = []

    return 1. / sum([x * x for x in all_scores])


def max_approval_score(election):
    score = np.zeros([election.num_candidates])
    for vote in election.votes:
        for c in vote:
            score[c] += 1
    return max(score)


def is_condorcet(election):
    """ Check if election witness Condorcet winner"""
    potes = election.get_potes() # get positional votes
    for i in range(election.num_candidates):

        condocret_winner = True
        for j in range(election.num_candidates):

            diff = 0
            for k in range(election.num_voters):

                if potes[k][i] <= potes[k][j]:
                    diff += 1

            if diff < math.ceil((election.num_voters + 1) / 2.):
                condocret_winner = False
                break

        if condocret_winner:
            return True

    return False



import numpy as np

from mapel.core.inner_distances import l2

from itertools import combinations

def map_diameter(c: int) -> float:
    """ Compute the diameter """
    return 1. / 3. * (c + 1) * (c - 1)

def distortion_from_guardians(experiment, election_id) -> np.ndarray:
    values = np.array([])
    election_id_1 = election_id

    for election_id_2 in experiment.elections:
        if election_id_2 in {'identity_10_100_0', 'uniformity_10_100_0',
                             'antagonism_10_100_0', 'stratification_10_100_0'}:
            if election_id_1 != election_id_2:
                m = experiment.elections[election_id_1].num_candidates
                true_distance = experiment.distances[election_id_1][election_id_2]
                true_distance /= map_diameter(m)
                embedded_distance = l2(experiment.coordinates[election_id_1],
                                       experiment.coordinates[election_id_2])

                embedded_distance /= \
                    l2(experiment.coordinates['identity_10_100_0'],
                       experiment.coordinates['uniformity_10_100_0'])
                ratio = float(true_distance) / float(embedded_distance)
                values = np.append(values, ratio)

    return values


def distortion_from_all(experiment, election_ids):

    all_values = {}

    for election_id in election_ids:

        values = np.array([])
        one_side_values = np.array([])
        election_id_1 = election_id

        for election_id_2 in experiment.instances:
            # if election_id_2 in {'identity_10_100_0', 'uniformity_10_100_0',
            #                      'antagonism_10_100_0', 'stratification_10_100_0'}:
            if election_id_1 != election_id_2:
                # m = election.instances[election_id_1].num_candidates
                true_distance = experiment.distances[election_id_1][election_id_2]
                true_distance /= experiment.distances['ID']['UN']
                embedded_distance = l2(np.array(experiment.coordinates[election_id_1]),
                                       np.array(experiment.coordinates[election_id_2]))

                embedded_distance /= \
                    l2(np.array(experiment.coordinates['ID']),
                       np.array(experiment.coordinates['UN']))
                # try:
                #     ratio = float(embedded_distance) / float(true_distance)
                # except:
                #     ratio = 1.
                one_side_ratio = embedded_distance / true_distance
                one_side_values = np.append(one_side_values, one_side_ratio)

                ratio = max(embedded_distance, true_distance) / min(embedded_distance, true_distance)

                values = np.append(values, ratio)

        all_values[election_id_1] = np.mean(values)

    return all_values


# def distortion_from_all(election, election_id) -> np.ndarray:
#     values = np.array([])
#     election_id_1 = election_id
#
#     for election_id_2 in election.elections:
#         if election_id_1 != election_id_2:
#             m = election.elections[election_id_1].num_candidates
#             true_distance = election.distances[election_id_1][election_id_2]
#             true_distance /= map_diameter(m)
#             embedded_distance = l2(np.array(election.coordinates[election_id_1]),
#                                    np.array(election.coordinates[election_id_2]))
#
#             embedded_distance /= \
#                 l2(np.array(election.coordinates['core_800']),
#                    np.array(election.coordinates['core_849']))
#             try:
#                 ratio = float(embedded_distance) / float(true_distance)
#             except:
#                 ratio = 1.
#             values = np.append(values, ratio)
#
#     return np.mean(abs(1.-values))


# def distortion_from_top_100(election, election_id) -> np.ndarray:
#     values = np.array([])
#     election_id_1 = election_id
#
#     euc_dist = {}
#     for election_id_2 in election.elections:
#         if election_id_1 != election_id_2:
#             euc_dist[election_id_2] = l2(np.array(election.coordinates[election_id_1]),
#                                            np.array(election.coordinates[election_id_2]))
#
#     all = (sorted(euc_dist.items(), key=lambda item: item[1]))
#     top_100 = [x for x,_ in all[0:100]]
#
#
#     # all = (sorted(election.distances[election_id_1].items(), key=lambda item: item[1]))
#     # top_100 = [x for x,_ in all[0:100]]
#
#     for election_id_2 in election.elections:
#         if election_id_1 != election_id_2:
#             if election_id_2 in top_100:
#                 m = election.elections[election_id_1].num_candidates
#                 true_distance = election.distances[election_id_1][election_id_2]
#                 true_distance /= map_diameter(m)
#                 embedded_distance = l2(np.array(election.coordinates[election_id_1]),
#                                        np.array(election.coordinates[election_id_2]))
#
#                 embedded_distance /= \
#                     l2(np.array(election.coordinates['core_800']),
#                        np.array(election.coordinates['core_849']))
#                 try:
#                     ratio = float(embedded_distance) / float(true_distance)
#                 except:
#                     ratio = 1.
#                 values = np.append(values, ratio)
#
#     return np.mean(abs(1.-values))


def avg_distortion_from_guardians(experiment, election_id):
    values = distortion_from_guardians(experiment, election_id)
    return np.mean(values)


def worst_distortion_from_guardians(experiment, election_id):
    values = distortion_from_guardians(experiment, election_id)
    return np.max(values)



def distortion(experiment, election) -> float:
    e0 = election.election_id
    c0 = np.array(experiment.coordinates[e0])
    distortion = 0
    for e1, e2 in combinations(experiment.elections, 2):
        if e1 != e0 and e2 != e0:
            original_d1 = experiment.distances[e0][e1]
            original_d2 = experiment.distances[e0][e2]
            original_proportion = original_d1 / original_d2
            embedded_d1 = np.linalg.norm(c0 - experiment.coordinates[e1], ord=2)
            embedded_d2 = np.linalg.norm(c0 - experiment.coordinates[e2], ord=2)
            embedded_proportion = embedded_d1 / embedded_d2
            _max = max(original_proportion, embedded_proportion)
            _min = min(original_proportion, embedded_proportion)
            distortion += _max / _min
    return distortion


def monotonicity_triplets(experiment, election) -> float:
    epsilon = 0.1
    e0 = election.election_id
    c0 = np.array(experiment.coordinates[e0])
    distortion = 0.
    ctr = 0.
    for e1, e2 in combinations(experiment.elections, 2):
        if e1 != e0 and e2 != e0:
            original_d1 = experiment.distances[e0][e1]
            original_d2 = experiment.distances[e0][e2]
            embedded_d1 = np.linalg.norm(c0 - experiment.coordinates[e1], ord=2)
            embedded_d2 = np.linalg.norm(c0 - experiment.coordinates[e2], ord=2)
            if (original_d1 < original_d2 and embedded_d1 > embedded_d2 * (1. + epsilon)) or \
                    (original_d2 < original_d1 and embedded_d2 > embedded_d1 * (1. + epsilon)):
                distortion += 1.
            ctr += 1.
    distortion /= ctr
    print(distortion)
    return distortion




import scipy.special

from mapel.elections.features.scores import get_cc_score, get_cc_dissat


def get_banzhaf_cc_score(election, committee_size=1):
    if election.fake:
        return {'value': None, 'dissat': None}

    winners = set()
    BASE = {}
    BINOM = {}

    for _ in range(committee_size):
        highest_c = 0
        highest_score = 0
        for c in range(election.num_candidates):
            if c in winners:
                continue
            candidate_score = sum([voter_score(BASE, BINOM, election, v, c, committee_size, winners)
                                   for v in range(election.num_voters)])
            if candidate_score > highest_score:
                highest_score = candidate_score
                highest_c = c
        winners.add(highest_c)

    return {'value': get_cc_score(election, winners), 'dissat': get_cc_dissat(election, winners)}


def voter_score(BASE, BINOM, election, v, c, committee_size, winners):

    potes = election.get_potes()
    m = election.num_candidates

    Wa = 0
    for i in range(potes[v][c]):
        if election.votes[v][i] in winners:
            Wa += 1
    Wb = len(winners) - Wa

    c_score = delta_c(BASE, BINOM, potes[v][c], Wa, Wb, m, committee_size)

    r = []
    for i in range(m):
        if potes[v][i] < potes[v][c]:
            r.append(0)
        else:
            r.append(1)

    d_score = sum(delta_d(BASE, BINOM, potes[v][d], Wa, Wb, m, committee_size, r[d])
                  for d in range(election.num_candidates) if d != c)

    return c_score + d_score


def delta_c(BASE, BINOM, pos, Wa, Wb, m, committee_size):
    name = f'c_{pos}_{Wa}_{Wb}'
    if name not in BASE:
        score = big_C(BINOM, pos, Wa, Wb, m, committee_size) * (m-pos-1)
        BASE[name] = score
    return BASE[name]


def delta_d(BASE, BINOM, pos, Wa, Wb, m, committee_size, r_d):
    name = f'd_{pos}_{Wa}_{Wb}_{r_d}'
    if name not in BASE:
        score = big_C(BINOM, pos, Wa, Wb, m, committee_size) * ( (m-pos-1) - (m-pos-1)*(1-r_d) )
        BASE[name] = score
    return BASE[name]


def big_C(BINOM, pos, Wa, Wb, m, committee_size):
    if Wa == 0:
        name = f'{pos}_{Wb}'
        if name not in BINOM:

            if m - pos - Wb <= 0 or 1 - committee_size - Wb <= 0:
                b_2 = 1
            else:
                b_2 = scipy.special.binom(m - pos - Wb, 1 - committee_size - Wb)

            BINOM[name] = b_2

        return BINOM[name]
    return 0



#!/usr/bin/env python
import math
import numpy as np
from mapel.elections.distances import lp as lp

def randomize(vector, num_winners):
    scores = [x for x, _ in vector]
    ranking = [x for _, x in vector]
    last_value = scores[num_winners-1]
    # LEFT
    left = num_winners - 2
    while left >= 0 and scores[left] == last_value:
        left -= 1
    left += 1
    # RIGHT
    right = num_winners
    while right < len(scores) and scores[right] == last_value:
        right += 1
    # print(left, right)
    if left < right:
        ranking[left:right] = np.random.choice(ranking[left:right], right - left,
                                               replace=False)
    # print(ranking)
    return ranking


def compute_sntv_winners(election=None, num_winners=1):
    """ Compute SNTV winners for a given election """
    scores = [0 for _ in range(election.num_candidates)]
    for vote in election.votes:
        scores[vote[0]] += 1
    candidates = [i for i in range(election.num_candidates)]
    results = sorted(zip(scores, candidates), reverse=True)
    ranking = randomize(results, num_winners)
    return ranking[0:num_winners]


def compute_borda_winners(election=None, num_winners=1):
    """ Compute Borda winners for a given election """

    scores = [0 for _ in range(election.num_candidates)]
    for vote in election.votes:
        for i in range(election.num_candidates):
            scores[vote[i]] += election.num_candidates - i - 1
    candidates = [i for i in range(election.num_candidates)]
    results = sorted(zip(scores, candidates), reverse=True)
    ranking = randomize(results, num_winners)
    return ranking[0:num_winners]


def compute_stv_winners(election=None, num_winners=1):
    """ Compute STV winners for a given election """

    winners = [] #[0] * params['orders']
    active = [True] * election.num_candidates

    droop_quota = math.floor(election.num_voters / (num_winners + 1.) ) + 1

    votes_on_1 = [0.] * election.num_candidates
    for i in range(election.num_voters):
        votes_on_1[election.votes[i][0]] += 1

    v_power = [1.] * election.num_voters


    while len(winners) + sum(active) > num_winners:

        ctr = election.num_candidates
        winner_id = 0
        while ctr > 0:
            if active[winner_id] and votes_on_1[winner_id] >= droop_quota:

                winners += [winner_id]

                total = 0
                for i in range(election.num_voters):
                    for j in range(election.num_candidates):
                        if active[election.votes[i][j]]:
                            if election.votes[i][j] == winner_id:
                                for k in range(j + 1, election.num_candidates):
                                    if active[election.votes[i][k]]:
                                        v_power[i] *= float(votes_on_1[winner_id] - droop_quota) / float(votes_on_1[winner_id])
                                        votes_on_1[election.votes[i][k]] += 1. * v_power[i]
                                        total += 1. * v_power[i]
                                        ctr = election.num_candidates
                                        break
                            break
                votes_on_1[winner_id] = 0
                active[winner_id] = False

            ctr -= 1
            winner_id += 1
            winner_id %= election.num_candidates

        loser_votes = droop_quota
        loser_id = 0
        for i in range(election.num_candidates):
            if active[i] and votes_on_1[i] < loser_votes:
                loser_votes = votes_on_1[i]
                loser_id = i

        votes_on_1[loser_id] = 0
        for i in range(election.num_voters):
            for j in range(election.num_candidates):
                if active[election.votes[i][j]]:
                    if election.votes[i][j] == loser_id:
                        for k in range(j+1, election.num_candidates):
                            if active[election.votes[i][k]]:
                                votes_on_1[election.votes[i][k]] += 1. * v_power[i]
                                break
                    break
        active[loser_id] = False


    for i in range(election.num_candidates):
        if active[i]:
            winners += [i]

    winners = sorted(winners)

    return winners

###

def get_borda_points(votes, num_voters, num_candidates):
    points = np.zeros([num_candidates])
    scoring = [1. for _ in range(num_candidates)]

    for i in range(len(scoring)):
        scoring[i] = len(scoring) - i - 1

    for i in range(num_voters):
        for j in range(num_candidates):
            points[int(votes[i][j])] += scoring[j]

    return points


###


def generate_winners(election=None, num_winners=1, ballot="ordinal", type=None, name=None):
    votes, num_voters, num_candidates = election.votes, election.num_voters, election.num_candidates
    params = {"orders": num_winners,
              "pure": True,
              "elections": 1,
              'candidates': num_candidates,
              'voters': num_voters}
    rule = {'type_id': type,
            'name': name,
            'length': num_candidates}
    winners, total_time = get_winners(params, votes, rule, ballot)
    return winners, total_time


def get_winners(params, votes, rule, ballot='ordinal'):
    if ballot == "ordinal":
        return get_ordinal_winners(params, votes, rule)
    elif ballot == "approval":
        return get_approval_winners(params, votes, rule)


def get_approval_winners(params, elections, rule):
    if rule['type_id'] == 'app_cc':
        all_winners = []
        for i in range(params['elections']):
            winners = get_winners_app_cc(params, elections['votes'][i], elections['candidates'][i])
            all_winners += winners
        return all_winners


# Need update
def get_ordinal_winners(params, votes, rule):
    if rule['type_id'] == 'scoring':
        scoring = get_rule(rule['name'], rule['length'])
        all_winners = []
        for i in range(params['elections']):
            winners = get_winners_scoring(params, votes[i], params['candidates'], scoring)
            all_winners += winners
        return all_winners

    elif rule['type_id'] == 'borda_owa':
        owa = get_rule(rule['name'], rule['length'])
        all_winners = []
        for i in range(params['elections']):
            winners, total_time = get_winners_borda_owa(params, votes, owa)
            all_winners += winners
        # print(obj_vaue)
        return all_winners, total_time

    elif rule['type_id'] == 'bloc_owa':
        owa = get_rule(rule['name'], rule['length'])
        # t_bloc = rule['special']
        t_bloc = params["orders"]
        all_winners = []
        for i in range(params['elections']):
            winners, total_time = get_winners_bloc_owa(params, votes, owa, t_bloc)
            all_winners += winners
        # print(winners)
        return all_winners, total_time

    elif rule['type_id'] == 'stv':
        all_winners = []
        for i in range(params['elections']):
            winners = get_winners_stv(params, votes[i], params['candidates'])
            all_winners += winners
        return all_winners

    elif rule['type_id'] == 'election':
        all_winners = []
        for i in range(params['elections']):
            winners = []
            for j in range(params['voters']):
                winners += [votes[i][j]]

            all_winners += winners
        return all_winners

    # elif rule['type_id'] == "approx_cc":
    #     all_winners = []
    #     winners = get_winners_approx_cc(votes, params)
    #     all_winners += winners
    #     return all_winners
    #
    # elif rule['type_id'] == "approx_hb":
    #     all_winners = []
    #     winners = get_winners_approx_hb(votes, params, 'greedy')
    #     all_winners += winners
    #     return all_winners
    #
    # elif rule['type_id'] == "approx_pav":
    #     all_winners = []
    #     winners = get_winners_approx_pav(votes, params, 'greedy')
    #     all_winners += winners
    #     return all_winners

def get_rule(name, length):

    rule = [0.] * length
    if name == "borda":
        for i in range(length):
            rule[i] = (length - float(i) - 1.) / (length - 1.)
    elif name == 'sntv':
        rule[0] = 1.
    elif name == 'cc':
        rule[0] = 1.
    elif name == 'hb':
        for i in range(length):
            rule[i] = 1. / (i + 1.)
    else:
        return name
    return rule


def get_winners_app_cc(params, votes, candidates):

    points = [0 for _ in range(params["candidates"])]

    for i in range(params['voters']):
        for j in range(len(votes[i])):
            points[int(votes[i][j])] += 1.

    tmp_candidates = [x for _, x in sorted(zip(points, candidates))]
    winners = tmp_candidates[params['candidates'] - params['orders']: params['candidates']]

    winners = sorted(winners)
    return winners


def get_winners_scoring(params, votes, candidates, scoring):

    points = [0 for _ in range(params["candidates"])]

    for i in range(params['voters']):
        for j in range(params['candidates']):
            points[int(votes[i][j])] += scoring[j]

    tmp_candidates = [x for _, x in sorted(zip(points, candidates))]
    winners = tmp_candidates[params['candidates'] - params['orders']: params['candidates']]

    points = sorted(points)
    ### DRAW ###
    if params['orders'] > 1:
        breaking_point = params['candidates'] - params['orders']
        if points[breaking_point-1] == points[breaking_point]:
            left = -1
            while points[breaking_point-1 + left] == points[breaking_point + left]:
                left -= 1
            right = 0

            while right < params['orders']-2 and points[breaking_point+right] == points[breaking_point+right+1]:
                right += 1
            while right >= 0:
                r = np.random.randint(left,right)
                tmp_candidate = winners[right]
                winners[right] = tmp_candidates[breaking_point+r]
                tmp_candidates[breaking_point + r] = tmp_candidate
                right -= 1
    winners = sorted(winners)
    return winners


def get_winners_borda_owa(params, votes, owa):
    return lp.solve_lp_borda_owa(params, votes, owa)


def get_winners_bloc_owa(params, votes, owa, t_bloc):
    return lp.solve_lp_bloc_owa(params, votes, owa, t_bloc)


def get_winners_stv(params, votes, candidates):

    winners = []
    active = [True] * params['candidates']

    droop_quota = math.floor(params['voters'] / (params['orders'] + 1.) ) + 1

    votes_on_1 = [0.] * params['candidates']
    for i in range(params['voters']):
        votes_on_1[votes[i][0]] += 1

    v_power = [1.] * params['voters']

    while len(winners) + sum(active) > params['orders']:

        ctr = params['candidates']
        winner_id = 0
        while ctr > 0:

            if active[winner_id] and votes_on_1[winner_id] >= droop_quota:

                if params['pure']:
                    winners += [winner_id]
                else:
                    winners += [candidates[winner_id]]


                total = 0
                for i in range(params['voters']):
                    for j in range(params['candidates']):
                        if active[votes[i][j]]:
                            if votes[i][j] == winner_id:
                                for k in range(j + 1, params['candidates']):
                                    if active[votes[i][k]]:
                                        v_power[i] *= float(votes_on_1[winner_id] - droop_quota) / float(votes_on_1[winner_id])
                                        votes_on_1[votes[i][k]] += 1. * v_power[i]
                                        total += 1. * v_power[i]
                                        ctr = params['candidates']
                                        break
                            break

                votes_on_1[winner_id] = 0
                active[winner_id] = False

            ctr -= 1
            winner_id += 1
            winner_id %= params['candidates']

        loser_votes = droop_quota
        loser_id = 0
        for i in range(params['candidates']):
            if active[i] and votes_on_1[i] < loser_votes:
                loser_votes = votes_on_1[i]
                loser_id = i

        votes_on_1[loser_id] = 0
        for i in range(params['voters']):
            for j in range(params['candidates']):
                if active[votes[i][j]]:
                    if votes[i][j] == loser_id:
                        for k in range(j+1, params['candidates']):
                            if active[votes[i][k]]:
                                votes_on_1[votes[i][k]] += 1. * v_power[i]
                                break
                    break
        active[loser_id] = False

    for i in range(params['candidates']):
        if active[i]:
            if params['pure']:
                winners += [i]
            else:
                winners += [candidates[i]]

    winners = sorted(winners)

    return winners





from numpy import ceil
import sys
import os

try:
    import pulp
except Exception:
    pulp = None

try:
    from dotenv import load_dotenv

    load_dotenv()
    sys.path.append(os.environ["PATH"])
    from abcvoting import abcrules, preferences
except ImportError:
    abcrules = None
    preferences = None


def calculate_committees(election, resolute=False,
                         committee_size: int = 10, rule_name=None) -> set:
    profile = preferences.Profile(num_cand=election.num_candidates)
    profile.add_voters(election.votes)
    try:
        committees = abcrules.compute(rule_name, profile, committee_size, resolute=resolute)
    except Exception:
        committees = {}

    return committees


def count_proportionality_degree_of_a_committee(election, committee: set,
                                                committee_size: int = 10) -> map:
    f_map = dict()
    for l in range(1, committee_size + 1):
        val = solve_ilp_instance(election, committee, l, committee_size=committee_size)
        f_map[l] = val
    return f_map


def solve_ilp_instance(election, committee: set, l: int = 1,
                       committee_size: int = 10) -> float:
    model = pulp.LpProblem("pd_value_f", pulp.LpMinimize)
    X = [pulp.LpVariable("x_" + str(i), cat='Binary') for i in
         range(election.num_voters)]  # X[i] = 1 if we select i-th voter, otherwise 0
    Y = [pulp.LpVariable("y_" + str(j), cat='Binary') for j in
         range(election.num_candidates)]  # Y[j] = 1 if we select j-th candidate, otherwise 0
    s = int(ceil(
        l * election.num_voters / committee_size))  # If there is any valid l-cohesive group, then there is also at least one with minimum possible size

    objective = 0
    for v in range(election.num_voters):
        cands = election.votes[v]
        for c in committee:
            if c in cands:
                objective += X[v]
    model += objective  # We want to minimize the number of votes

    x_sum_eq = 0
    for x in X:
        x_sum_eq += x
    model += x_sum_eq == s  # We choose exactly s voters

    y_sum_ineq = 0
    for y in Y:
        y_sum_ineq += y
    model += y_sum_ineq >= l  # We choose at least l candidates (although l are sufficient in this case)

    cand_to_voters_variables_list = [[] for j in range(election.num_candidates)]
    for i, d in enumerate(election.votes):
        for j in d:
            cand_to_voters_variables_list[j].append(X[i])
    # We want to assert that the selected voters approve all the selected candidates.
    # For each candidate j,  we construct the following inequality:  a_{0,j} * x_0 + a_{1,j} * x_1 + ... + a_{n-1,j} * x_{n-1}  -   s * y_j    >=    0
    # We define a_{i, j} as the flag indicating whether i-th voter approves j-th candidate (1 if yes, otherwise 0)
    # Let us observe that if the j-th candidate is not selected, then s * y_j = 0 and the above inequality is naturally satisfied.
    # However, if j-th candidate is selected, then the above can be satisfied if and only if all s selected voters approve j-th candidate
    for j, y in enumerate(Y):
        y_ineq = 0
        for x in cand_to_voters_variables_list[j]:
            y_ineq += x
        y_ineq -= s * y
        model += y_ineq >= 0

    # culture_id.solve()

    model.solve(pulp.PULP_CBC_CMD(msg=False))
    # if LpStatus[culture_id.status] == 'Optimal':
    #     print([var.election_id + "=" + str(var.varValue) for var in culture_id.variables() if var.varValue is not None and var.varValue > 0], sep=" ")    # prints result variables which have value > 0
    if pulp.LpStatus[model.status] == 'Optimal':
        return pulp.value(model.objective) / s
    else:
        # return np.inf
        return 0.


def proportionality_degree(election, committee_size=10, rule_name=None, resolute=False):
    committees = calculate_committees(election, committee_size=committee_size, rule_name=rule_name,
                                      resolute=resolute)

    if len(committees) == 0:
        return 0

    all_pd = []
    for committee in committees:
        pd_1 = solve_ilp_instance(election, committee, 1, committee_size=committee_size)
        all_pd.append(pd_1)
        # print(f"Election with n = {election.num_voters}, m = {election.num_candidates}, k = {election.k}.   Votes = {election.votes}")
        # print(f"PAV committee = {committee}.    Proportionality Degree of this committee is {pd}.\n")
    return sum(all_pd) / len(all_pd)


def proportionality_degree_av(*args, **kwargs):
    return proportionality_degree(*args, **kwargs, rule_name='av')


def proportionality_degree_pav(*args, **kwargs):
    return proportionality_degree(*args, **kwargs, rule_name='pav')


def proportionality_degree_cc(*args, **kwargs):
    return proportionality_degree(*args, **kwargs, rule_name='cc', resolute=True)



"""
This parser is compatible with Preflib file format in October 2023
Sample code for using it:
from mapel.elections.other.parser import preflib_parser
votes, nr_candidates, nr_votes, label = preflib_parser(file_name="~Input the absolute path to the file~", file_ending=4)
"""

import re
import numpy as np

regex_file_name = r'# FILE NAME:'
regex_title = r'# TITLE:'
regex_data_type = r'# DATA TYPE:'
regex_number_alternatives = r"# NUMBER ALTERNATIVES:"
regex_number_voters = r"# NUMBER VOTERS:"
regex_number_unique_orders = r"# NUMBER UNIQUE ORDERS:"
regex_number_categories = r"# NUMBER CATEGORIES:"


def validate():
    pass


def process_soc_line(line: str, votes: list):
    tokens = line.split(':')
    nr_this_vote = int(tokens[0])
    vote = [int(x) for x in tokens[1].split(',')]
    vote = np.array([x-1 for x in vote])
    for i in range(0, nr_this_vote):
        votes.append(vote)
    pass


def process_soi_line(line: str, votes: list):
    pass


def process_toc_line(line: str, votes: list):
    pass


def process_toi_line(line: str, votes: list):
    pass


def preflib_parser(file_name: str, label: str = None, file_ending: int = 4):
    votes = []
    nr_candidates = 0
    nr_votes = 0
    nr_unique = 0
    alternative_names = list()
    from_file_file_name = ''
    from_file_title = ''
    from_file_data_type = ''
    if label == "" or label is None:
        label = file_name[:-file_ending]
    file = open(file_name, "r")
    # read metadata
    for line in file:
        if line[-1] == '\n':
            line = line[:-1]
        if line[0] != '#':
            if from_file_data_type == 'soc':
                process_soc_line(line, votes)
            elif from_file_data_type == 'soi':
                process_soi_line(line, votes)
            elif from_file_data_type == 'toc':
                process_toc_line(line, votes)
            elif from_file_data_type == 'toi':
                process_toi_line(line, votes)
            break
        elif re.search(regex_file_name, line):
            from_file_file_name = line.split(':')[1][1:-file_ending]
        elif re.search(regex_title, line):
            from_file_title = line.split(':')[1].replace(" ", "")
        elif re.search(regex_data_type, line):
            from_file_data_type = line.split(':')[1][1:]
        elif re.search(regex_number_alternatives, line):
            nr_candidates = int(line.split(':')[1])
        elif re.search(regex_number_voters, line):
            nr_votes = int(line.split(':')[1])
        elif re.search(regex_number_unique_orders, line):
            nr_unique = int(line.split(':')[1])

    label = from_file_title + "_" + from_file_file_name
    # read votes
    if from_file_data_type == 'soc':
        for line in file:
            process_soc_line(line, votes)
    elif from_file_data_type == 'soi':
        for line in file:
            process_soi_line(line, votes)
    elif from_file_data_type == 'toc':
        for line in file:
            process_toc_line(line, votes)
    elif from_file_data_type == 'toi':
        for line in file:
            process_toi_line(line, votes)
    else:
        print(label, " - unknown data format.")

    return np.array(votes), nr_candidates, nr_votes, label
